from pharmpy.model import Model
from pharmpy.workflows import Task, Workflow


def create_workflow(models=None, n=None, tool=None):
    """
    If models is None and n is None: fit one unknown model
    If models is model object (n must be None): fit that one model
    If models is list of models (n must be None): fit these models
    If models is None and n is an int: fit n unknown models
    """
    wf = create_fit_workflow(models, n, tool)
    wf.name = "modelfit"
    task_result = Task('results', post_process_results)
    wf.add_task(task_result, predecessors=wf.output_tasks)
    return wf


def create_fit_workflow(models=None, n=None, tool=None):
    from pharmpy.tools.modelfit import conf

    if tool is None:
        tool = conf.default_tool

    if tool == 'nonmem':
        from pharmpy.plugins.nonmem.run import execute_model
    elif tool == 'nlmixr':
        from pharmpy.plugins.nlmixr.run import execute_model
    else:
        raise ValueError(f"Unknown estimation tool {tool}")

    wf = Workflow()
    if models is None:
        if n is None:
            task = Task('run', execute_model)
            wf.add_task(task)
        else:
            for i in range(n):
                task = Task(f'run{i}', execute_model)
                wf.add_task(task)
    elif isinstance(models, Model):
        task = Task('run', execute_model, models)
        wf.add_task(task)
    else:
        for i, model in enumerate(models):
            task = Task(f'run{i}', execute_model, model)
            wf.add_task(task)
    return wf


def post_process_results(*models):
    if len(models) > 1:
        return models
    else:
        return models[0]



from itertools import combinations

import pharmpy.tools.modelfit as modelfit
from pharmpy.modeling import copy_model, remove_iiv
from pharmpy.modeling.block_rvs import create_joint_distribution, split_joint_distribution
from pharmpy.workflows import Task, Workflow


def brute_force_no_of_etas(iivs):
    wf = Workflow()
    model_features = dict()

    eta_names = [eta.name for eta in iivs]
    eta_combos = _get_combinations(eta_names, include_single=True)

    for i, combo in enumerate(eta_combos, 1):
        model_name = f'iiv_no_of_etas_candidate{i}'
        task_copy = Task('copy', copy, model_name)
        wf.add_task(task_copy)

        task_remove_eta = Task('remove_eta', remove_eta, combo)
        wf.add_task(task_remove_eta, predecessors=task_copy)

        model_features[model_name] = combo

    wf_fit = modelfit.create_workflow(n=i)
    wf.insert_workflow(wf_fit)
    return wf, model_features


def remove_eta(etas, model):
    remove_iiv(model, etas)
    return model


def brute_force_block_structure(iivs):
    wf = Workflow()
    model_features = dict()

    eta_combos_single_blocks, eta_combos_multi_blocks = _get_possible_iiv_blocks(iivs)
    eta_combos_all = [None] + eta_combos_single_blocks + eta_combos_multi_blocks

    model_no = 1
    for combo in eta_combos_all:
        # Do not run model with same block structure as start model
        if combo:
            if _is_current_block_structure(iivs, combo):
                continue
        else:
            if all(len(rv.joint_names) == 0 for rv in iivs):
                continue

        model_name = f'iiv_block_structure_candidate{model_no}'
        task_copy = Task('copy', copy, model_name)
        wf.add_task(task_copy)

        if combo is None:
            task_joint_dist = Task('split_joint_dist', split_joint_dist)
            wf.add_task(task_joint_dist, predecessors=task_copy)
            model_features[model_name] = [[eta.name] for eta in iivs]
        else:
            task_joint_dist = Task('create_joint_dist', create_joint_dist, combo)
            wf.add_task(task_joint_dist, predecessors=task_copy)
            model_features[model_name] = combo
        model_no += 1
    wf_fit = modelfit.create_workflow(n=len(model_features))
    wf.insert_workflow(wf_fit)
    return wf, model_features


def _is_current_block_structure(iivs, list_of_etas):
    if not isinstance(list_of_etas[0], list):
        list_of_etas = [list_of_etas]

    for eta_block in list_of_etas:
        eta_name = eta_block[0]
        if iivs[eta_name].joint_names != eta_block:
            return False
    return True


def _get_combinations(names, include_single=False):
    combos = []
    if include_single:
        start = 1
    else:
        start = 2
    for i in range(start, len(names) + 1):
        combos += [list(combo) for combo in combinations(names, i)]
    return combos


def _get_possible_iiv_blocks(iivs):
    eta_names = iivs.names
    eta_combos_single_blocks = _get_combinations(eta_names)
    if len(eta_names) < 4:
        return eta_combos_single_blocks, []

    no_of_blocks_max = int(len(eta_names) / 2)
    eta_combos_multi_blocks = []

    for i in range(2, no_of_blocks_max + 1):
        etas = []
        for combo in combinations(eta_combos_single_blocks, i):
            combo = list(combo)
            no_of_etas_in_combo = len(flatten(combo))
            no_of_etas_unique = len(set(flatten(combo)))
            if no_of_etas_in_combo == len(eta_names) and no_of_etas_unique == len(eta_names):
                etas.append(combo)
        eta_combos_multi_blocks += etas

    return eta_combos_single_blocks, eta_combos_multi_blocks


def flatten(list_to_flatten):
    return [item for sublist in list_to_flatten for item in sublist]


def create_joint_dist(list_of_etas, model):
    if isinstance(list_of_etas[0], list):
        for eta_block in list_of_etas:
            create_joint_distribution(model, eta_block)
    else:
        create_joint_distribution(model, list_of_etas)
    return model


def split_joint_dist(model):
    split_joint_distribution(model)
    return model


def copy(name, model):
    model_copy = copy_model(model, name)
    return model_copy



"""DataInfo is a companion to the dataset. It contains metadata of the dataset
"""
import copy
import json
from collections.abc import MutableSequence
from pathlib import Path

import pandas as pd
import sympy

from pharmpy.utils import parse_units


class ColumnInfo:
    """Information about one data column

    Parameters
    ----------
    name : str
        Colum name
    type : str
        Type (see the "type" attribute)
    unit : str
        Unit (see the "unit" attribute)
    scale : str
        Scale of measurement (see the "scale" attribute)
    continuous : bool
        True if continuous or False if discrete
    categories : list
        List of all possible categories
    drop : bool
        Should column be dropped (i.e. barred from being used)
    datatype : str
        Pandas datatype or special Pharmpy datatype (see the "dtype" attribute)
    descriptor : str
        Descriptor (kind) of data
    """

    _all_types = [
        'id',
        'dv',
        'idv',
        'unknown',
        'dose',
        'additional',
        'ii',
        'event',
        'covariate',
        'mdv',
        'nmtran_date',
    ]
    _all_scales = ['nominal', 'ordinal', 'interval', 'ratio']
    _all_dtypes = [
        'int8',
        'int16',
        'int32',
        'int64',
        'uint8',
        'uint16',
        'uint32',
        'uint64',
        'float16',
        'float32',
        'float64',
        'float128',
        'nmtran-time',
        'nmtran-date',
    ]
    _all_descriptors = [None, 'age', 'body weight', 'lean body mass', 'fat free mass']

    def __init__(
        self,
        name,
        type='unknown',
        unit=sympy.Integer(1),
        scale='ratio',
        continuous=True,
        categories=None,
        drop=False,
        datatype="float64",
        descriptor=None,
    ):
        if scale in ['nominal', 'ordinal']:
            continuous = False
        self._continuous = continuous
        self.name = name
        self.type = type
        self.unit = unit
        self.scale = scale
        self.continuous = continuous
        self.categories = categories  # dict from value to descriptive string
        self.drop = drop
        self.datatype = datatype
        self.descriptor = descriptor

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.type == other.type
            and self.unit == other.unit
            and self.scale == other.scale
            and self.continuous == other.continuous
            and self.categories == other.categories
            and self.drop == other.drop
            and self.datatype == other.datatype
        )

    @property
    def name(self):
        """Column name"""
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise TypeError("Column name must be a string")
        self._name = value

    @property
    def type(self):
        """Type of column

        ============  =============
        type          Description
        ============  =============
        id            Individual identifier. Max one per DataFrame. All values have to be unique
        idv           Independent variable. Max one per DataFrame.
        dv            Dependent variable
        covariate     Covariate
        dose          Dose amount
        additional    Number of additional doses
        ii            Interdose interval
        event         0 = observation
        mdv           0 = DV is observation value, 1 = DV is missing
        unknown       Unkown type. This will be the default for columns that hasn't been
                      assigned a type
        ============  =============
        """
        return self._type

    @type.setter
    def type(self, value):
        if value not in ColumnInfo._all_types:
            raise TypeError(f"Unknown column type {value}")
        self._type = value

    @property
    def descriptor(self):
        """Kind of data

        ================ =================
        descriptor       Description
        ================ =================
        age              Age (since birth)
        body weight      Human body weight
        lean body mass   Lean body mass
        fat free mass    Fat free mass
        ================ =================
        """
        return self._descriptor

    @descriptor.setter
    def descriptor(self, value):
        if value not in ColumnInfo._all_descriptors:
            raise TypeError(f"Unknown column descriptor {value}")
        self._descriptor = value

    @property
    def unit(self):
        """Unit of the column data

        Custom units are allowed, but units that are available in sympy.physics.units can be
        recognized. The default unit is 1, i.e. without unit.
        """
        return self._unit

    @unit.setter
    def unit(self, value):
        a = parse_units(value)
        self._unit = a

    @property
    def scale(self):
        """Scale of measurement

        The statistical scale of measurement for the column data. Can be one of
        'nominal', 'ordinal', 'interval' and 'rational'.
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        if value not in ColumnInfo._all_scales:
            raise TypeError(
                f"Unknown scale of measurement {value}. Only {ColumnInfo._all_scales} are possible."
            )
        self._scale = value
        if self.continuous and value in ['nominal', 'ordinal']:
            self.continuous = False

    @property
    def continuous(self):
        """Is the column data continuous

        True for continuous data and False for discrete. Note that nominal and ordinal data have to
        be discrete.
        """
        return self._continuous

    @continuous.setter
    def continuous(self, value):
        if value and self.is_categorical():
            raise ValueError(
                f"Cannot set variable on {self.scale} scale of measurement to continuous"
            )
        self._continuous = value

    @property
    def categories(self):
        """List of allowed categories"""
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

    @property
    def drop(self):
        """Should this column be dropped"""
        return self._drop

    @drop.setter
    def drop(self, value):
        self._drop = bool(value)

    @property
    def datatype(self):
        """Column datatype

        ============ ================ ======== ================================= ===========
        datatype     Description      Size     Range                             NA allowed?
        ============ ================ ======== ================================= ===========
        int8         Signed integer   8 bits   -128 to +127.                     No
        int16        Signed integer   16 bits  -32,768 to +32,767.               No
        int32        Signed integer   32 bits  -2,147,483,648 to +2,147,483,647. No
        int64        Signed integer   64 bits  -9,223,372,036,854,775,808 to     No
                                               9,223,372,036,854,775,807.
        uint8        Unsigned integer 8 bits   0 to 256.                         No
        uint16       Unsigned integer 16 bit   0 to 65,535.                      No
        uint32       Unsigned integer 32 bit   0 to 4,294,967,295.               No
        uint64       Unsigned integer 64 bit   0 to 18,446,744,073,709,551,615   No
        float16      Binary float     16 bits  ≈ ±6.55×10⁴                       Yes
        float32      Binary float     32 bits  ≈ ±3.4×10³⁸                       Yes
        float64      Binary float     64 bits  ≈ ±1.8×10³⁰⁸                      Yes
        float128     Binary float     128 bits ≈ ±1.2×10⁴⁹³²                     Yes
        nmtran-time  NM-TRAN time     n                                          No
        nmtran-date  NM-TRAN date     n                                          No
        ============ ================ ========================================== ===========

        The default, and most common datatype, is float64.
        """
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        if value not in ColumnInfo._all_dtypes:
            raise ValueError(
                f"{value} is not a valid datatype. Valid datatypes are {ColumnInfo._all_dtypes}"
            )
        self._datatype = value

    def is_categorical(self):
        """Check if the column data is categorical

        Returns
        -------
        bool
            True if categorical (nominal or ordinal) and False otherwise.

        See also
        --------
        is_numerical : Check if the column data is numerical

        Examples
        --------
        >>> from pharmpy.datainfo import ColumnInfo
        >>> col1 = ColumnInfo("WGT", scale='ratio')
        >>> col1.is_categorical()
        False
        >>> col2 = ColumnInfo("ID", scale='nominal')
        >>> col2.is_categorical()
        True

        """
        return self.scale in ['nominal', 'ordinal']

    def is_numerical(self):
        """Check if the column data is numerical

        Returns
        -------
        bool
            True if numerical (interval or ratio) and False otherwise.

        See also
        --------
        is_categorical : Check if the column data is categorical

        Examples
        --------
        >>> from pharmpy.datainfo import ColumnInfo
        >>> col1 = ColumnInfo("WGT", scale='ratio')
        >>> col1.is_numerical()
        True
        >>> col2 = ColumnInfo("ID", scale='nominal')
        >>> col2.is_numerical()
        False

        """
        return self.scale in ['interval', 'ratio']

    def copy(self):
        """Create a deep copy of the ColumnInfo

        Returns
        -------
        ColumnInfo
            Copied object
        """
        return copy.deepcopy(self)

    def __repr__(self):
        di = DataInfo([self])
        return repr(di)


class DataInfo(MutableSequence):
    """Metadata for the dataset

    Can be indexed to get ColumnInfo for the columns.

    Parameters
    ----------
    columns : list
        List of column names
    path : Path
        Path to dataset file
    separator : str
        Character or regexp separator for dataset
    """

    def __init__(self, columns=None, path=None, separator=','):
        if columns is None:
            self._columns = []
        elif len(columns) > 0 and isinstance(columns[0], str):
            self._columns = []
            for name in columns:
                colinf = ColumnInfo(name)
                self._columns.append(colinf)
        else:
            self._columns = columns
        self.path = path
        self.separator = separator

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for col1, col2 in zip(self, other):
            if col1 != col2:
                return False
        return self.path == other.path

    def __len__(self):
        return len(self._columns)

    def _getindex(self, i):
        if isinstance(i, str):
            for n, col in enumerate(self._columns):
                if col.name == i:
                    return n
            raise IndexError(f"Cannot find column {i} in DataInfo")
        elif isinstance(i, int):
            return i
        else:
            raise TypeError(f"Cannot index DataInfo by {type(i)}")

    def __getitem__(self, i):
        if isinstance(i, list):
            cols = []
            for ind in i:
                index = self._getindex(ind)
                cols.append(self._columns[index])
            return DataInfo(columns=cols)
        return self._columns[self._getindex(i)]

    def __setitem__(self, i, value):
        self._columns[self._getindex(i)] = value

    def __delitem__(self, i):
        del self._columns[self._getindex(i)]

    def insert(self, i, value):
        self._columns.insert(self._getindex(i), value)

    @property
    def path(self):
        """Path of dataset file

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.path     # doctest: +ELLIPSIS
        ...pharmpy/modeling/example_models/pheno.dta')
        """
        return self._path

    @path.setter
    def path(self, value):
        if value is not None:
            self._path = Path(value)
        else:
            self._path = None

    @property
    def separator(self):
        """Separator for dataset file

        Can be a single character or a regular expression
        string.
        """
        return self._separator

    @separator.setter
    def separator(self, value):
        self._separator = value

    @property
    def typeix(self):
        """Type indexer

        Example
        -------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.typeix['covariate'].names
        ['WGT', 'APGR']
        """
        return TypeIndexer(self)

    @property
    def id_column(self):
        """The id column

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.id_column.name
        'ID'
        """
        return self.typeix['id'][0]

    @id_column.setter
    def id_column(self, value):
        self[value].type = 'id'

    @property
    def dv_column(self):
        """The dv column

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.dv_column.name
        'DV'
        """
        return self.typeix['dv'][0]

    @dv_column.setter
    def dv_column(self, value):
        self[value].type = 'dv'

    @property
    def idv_column(self):
        """The idv column

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.idv_column.name
        'TIME'
        """
        return self.typeix['idv'][0]

    @idv_column.setter
    def idv_column(self, value):
        self[value].type = 'idv'

    @property
    def names(self):
        """All column names

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.names
        ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV']
        """
        return [col.name for col in self._columns]

    @property
    def types(self):
        """All column types"""
        return [col.type for col in self._columns]

    @types.setter
    def types(self, value):
        if isinstance(value, str):
            value = [value]
        if len(value) == 1:
            value *= len(self)
        if len(value) != len(self):
            raise ValueError(
                "Length mismatch. "
                "Can only set the same number of names as columns or 1 for broadcasting"
            )
        for v, col in zip(value, self._columns):
            col.type = v

    def to_json(self, path=None):
        a = []
        for col in self._columns:
            d = {
                "name": col.name,
                "type": col.type,
                "scale": col.scale,
                "continuous": col.continuous,
                "categories": col.categories,
                "unit": str(col.unit),
            }
            if col.descriptor is not None:
                d["descriptor"] = col.descriptor
            a.append(d)
        s = json.dumps({"columns": a})
        if path is None:
            return s
        else:
            with open(path, 'w') as fp:
                fp.write(s)

    @staticmethod
    def from_json(s):
        """Create DataInfo from JSON string

        Parameters
        ----------
        s : str
            JSON string

        Return
        ------
        DataInfo
            Created DataInfo object
        """
        d = json.loads(s)
        columns = []
        for col in d['columns']:
            ci = ColumnInfo(
                name=col['name'],
                type=col['type'],
                scale=col['scale'],
                continuous=col.get('continuous', True),
                unit=col.get('unit', sympy.Integer(1)),
                categories=col.get('categories', None),
                datatype=col.get('datatype', 'float64'),
                descriptor=col.get('descriptor', None),
            )
            columns.append(ci)
        return DataInfo(columns)

    @staticmethod
    def read_json(path):
        """Read DataInfo from JSON file

        Parameters
        ----------
        path : Path or str
            Path to JSON datainfo file

        Return
        ------
        DataInfo
            Created DataInfo object
        """
        with open(path, 'r') as fp:
            s = fp.read()
        return DataInfo.from_json(s)

    def copy(self):
        """Create a deep copy of the datainfo

        Returns
        -------
        DataInfo
            Copied object
        """
        return copy.deepcopy(self)

    def __repr__(self):
        labels = [col.name for col in self._columns]
        types = [col.type for col in self._columns]
        scales = [col.scale for col in self._columns]
        cont = [col.continuous for col in self._columns]
        cats = [col.categories for col in self._columns]
        units = [col.unit for col in self._columns]
        drop = [col.drop for col in self._columns]
        datatype = [col.datatype for col in self._columns]
        descriptor = [col.descriptor for col in self._columns]
        df = pd.DataFrame(columns=labels)
        df.loc['type'] = types
        df.loc['scale'] = scales
        df.loc['continuous'] = cont
        df.loc['categories'] = cats
        df.loc['unit'] = units
        df.loc['drop'] = drop
        df.loc['datatype'] = datatype
        df.loc['descriptor'] = descriptor
        return df.to_string()


class TypeIndexer:
    def __init__(self, obj):
        self._obj = obj

    def __getitem__(self, i):
        cols = [col for col in self._obj if col.type == i]
        if not cols:
            raise IndexError(f"No columns of type {i} available")
        return DataInfo(cols)



import os


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.githubpages',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_automodapi.automodapi',
    'jupyter_sphinx',
    'sphinxcontrib.autoprogram',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'Pharmpy'
year = '2018-2022'
authors = ['the Pharmpy development team']
copyright = '{0}; {1}'.format(year, ', '.join(authors))
version = release = '0.60.0'
html_show_sourcelink = False

pygments_style = 'trac'
templates_path = ['.']

html_static_path = ['.']
html_theme = "pydata_sphinx_theme"
html_css_files = [
    'custom.css',
]
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_theme_options = {
    'globaltoc_maxdepth': 2
}
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

automodapi_inheritance_diagram = False
graphviz_output_format = 'svg'

import doctest
doctest_default_flags = doctest.NORMALIZE_WHITESPACE
doctest_global_setup = '''
import pandas as pd
pd.set_option('display.width', 1000)
pd.set_option('display.max_columns', 1000)
'''



from abc import ABC, abstractmethod


class ModelDatabase(ABC):
    """Baseclass for databases for models and results of model runs

    Primary key is the model name, which means that models with
    the same name will be counted as the same model.

    Currently the model database is centered around storing and retrieving
    files belonging to the model or results from fitting the model. This
    doesn't mean that implementations have to use a file system to store the
    files.
    """

    @abstractmethod
    def store_local_file(self, model, path):
        """Store a file from the local machine

        Parameters
        ----------
        model : Model
            Pharmpy model object
        path : Path
            Path to file
        """
        pass

    @abstractmethod
    def retrieve_local_files(self, name, destination_path):
        """Retrieve all files related to a model

        Parameters
        ----------
        name : str
            Name of the model
        destination_path : Path
            Local destination path
        """
        pass

    @abstractmethod
    def retrieve_file(self, name, filename):
        """Retrieve one file related to a model

        Note that if the database is implemented in the local filesystem
        it is allowed to return a path directly to the file in the database.

        Parameters
        ----------
        name : str
            Name of the model
        filename : str
            Name of file

        Returns
        -------
        Path
            Path to local file
        """
        pass

    @abstractmethod
    def get_model(self, name):
        """Read a model from the database

        Parameters
        ----------
        name : str
            Name of the model

        Returns
        -------
        Model
            Retrieved model object
        """
        pass



import sympy

from pharmpy.statements import Assignment, sympify


def get_observation_expression(model):
    """Get the full symbolic expression for the observation according to the model

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_observation_expression
    >>> import sympy
    >>> model = load_example_model("pheno_linear")
    >>> expr = get_observation_expression(model)
    >>> sympy.pprint(expr)
    D_EPSETA1_2⋅EPS(1)⋅(ETA(2) - OETA₂) + D_ETA1⋅(ETA(1) - OETA₁) + D_ETA2⋅(ETA(2)
     - OETA₂) + EPS(1)⋅(D_EPS1 + D_EPSETA1_1⋅(ETA(1) - OETA₁)) + OPRED
    """
    stats = model.statements
    for i, s in enumerate(stats):
        if s.symbol == model.dependent_variable:
            y = s.expression
            break

    for j in range(i, -1, -1):
        y = y.subs({stats[j].symbol: stats[j].expression})

    return y


def get_individual_prediction_expression(model):
    """Get the full symbolic expression for the modelled individual prediction

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_individual_prediction_expression
    >>> model = load_example_model("pheno_linear")
    >>> get_individual_prediction_expression(model)
    D_ETA1*(ETA(1) - OETA1) + D_ETA2*(ETA(2) - OETA2) + OPRED

    See Also
    --------
    get_population_prediction_expression : Get full symbolic epression for the population prediction
    """
    y = get_observation_expression(model)
    for eps in model.random_variables.epsilons:
        y = y.subs({eps.symbol: 0})
    return y


def get_population_prediction_expression(model):
    """Get the full symbolic expression for the modelled population prediction

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_population_prediction_expression
    >>> model = load_example_model("pheno_linear")
    >>> get_population_prediction_expression(model)
    -D_ETA1*OETA1 - D_ETA2*OETA2 + OPRED

    See also
    --------
    get_individual_prediction_expression : Get full symbolic epression for the individual prediction
    """

    y = get_individual_prediction_expression(model)
    for eta in model.random_variables.etas:
        y = y.subs({eta.symbol: 0})
    return y


def calculate_eta_gradient_expression(model):
    """Calculate the symbolic expression for the eta gradient

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, calculate_eta_gradient_expression
    >>> model = load_example_model("pheno_linear")
    >>> calculate_eta_gradient_expression(model)
    [D_ETA1, D_ETA2]

    See also
    --------
    calculate_epsilon_gradient_expression : Epsilon gradient
    """

    y = get_observation_expression(model)
    for eps in model.random_variables.epsilons:
        y = y.subs({eps.symbol: 0})
    d = [y.diff(x.symbol) for x in model.random_variables.etas]
    return d


def calculate_epsilon_gradient_expression(model):
    """Calculate the symbolic expression for the epsilon gradient

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, calculate_epsilon_gradient_expression
    >>> model = load_example_model("pheno_linear")
    >>> calculate_epsilon_gradient_expression(model)
    [D_EPS1 + D_EPSETA1_1*(ETA(1) - OETA1) + D_EPSETA1_2*(ETA(2) - OETA2)]

    See also
    --------
    calculate_eta_gradient_expression : Eta gradient
    """

    y = get_observation_expression(model)
    d = [y.diff(x.symbol) for x in model.random_variables.epsilons]
    return d


def create_symbol(model, stem, force_numbering=False):
    """Create a new unique variable symbol given a model

    Parameters
    ----------
    model : Model
        Pharmpy model object
    stem : str
        First part of the new variable name
    force_numbering : bool
        Forces addition of number to name even if variable does not exist, e.g.
        COVEFF --> COVEFF1

    Returns
    -------
    Symbol
        Created symbol with unique name

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_symbol
    >>> model = load_example_model("pheno")
    >>> create_symbol(model, "TEMP")
    TEMP
    >>> create_symbol(model, "TEMP", force_numbering=True)
    TEMP1
    >>> create_symbol(model, "CL")
    CL1
    """
    symbols = [str(symbol) for symbol in model.statements.free_symbols]
    params = [param.name for param in model.parameters]
    rvs = [rv.name for rv in model.random_variables]
    dataset_col = model.datainfo.names
    misc = [model.dependent_variable]

    all_names = symbols + params + rvs + dataset_col + misc

    if str(stem) not in all_names and not force_numbering:
        return sympy.Symbol(str(stem))

    i = 1
    while True:
        candidate = f'{stem}{i}'
        if candidate not in all_names:
            return sympy.Symbol(candidate)
        i += 1


def _find_eta_assignments(model):
    # Is this find individual parameters?
    statements = model.statements.before_odes
    etas = {eta.symbol for eta in model.random_variables.etas}
    found = set()
    leafs = []
    for s in reversed(statements):
        if (
            etas & s.free_symbols
            and len(etas & statements.full_expression(s.symbol).free_symbols) == 1
            and s.symbol not in found
        ):
            leafs = [s] + leafs
            found.update(s.free_symbols)
    return leafs


def mu_reference_model(model):
    r"""Convert model to use mu-referencing

    Mu-referencing an eta is to separately define its actual mu (mean) parameter.
    For example: :math:`CL = \theta_1 e^{\eta_1}` with :math:`\eta_1` following a zero-mean
    normal distribution would give :math:`\mu_1 = \log{\theta_1}` and
    :math:`CL = e^{\mu_1 + \eta_1}`

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Model
        Reference to same object

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, mu_reference_model
    >>> model = load_example_model("pheno")
    >>> mu_reference_model(model).statements.before_odes
             ⎧TIME  for AMT > 0
             ⎨
    BTIME := ⎩ 0     otherwise
    TAD := -BTIME + TIME
    TVCL := THETA(1)⋅WGT
    TVV := THETA(2)⋅WGT
           ⎧TVV⋅(THETA(3) + 1)  for APGR < 5
           ⎨
    TVV := ⎩       TVV           otherwise
    μ₁ := log(TVCL)
           ETA(1) + μ₁
    CL := ℯ
    μ₂ := log(TVV)
          ETA(2) + μ₂
    V := ℯ
    S₁ := V
    """
    statements = model.statements.before_odes
    assignments = _find_eta_assignments(model)
    for i, eta in enumerate(model.random_variables.etas, start=1):
        for s in assignments:
            if eta.symbol in s.expression.free_symbols:
                assignment = statements.find_assignment(s.symbol)
                expr = assignment.expression
                indep, dep = expr.as_independent(eta.symbol)
                mu = sympy.Symbol(f'mu_{i}')
                newdep = dep.subs({eta.symbol: mu + eta.symbol})
                mu_expr = sympy.solve(expr - newdep, mu)[0]
                mu_ass = Assignment(mu, mu_expr)
                model.statements.insert_before(assignment, mu_ass)
                assignment.expression = newdep
    return model


def simplify_expression(model, expr):
    """Simplify expression given constraints in model

    Parameters
    ----------
    model : Model
        Pharmpy model object
    expr : Expression
        Expression to simplify

    Returns
    -------
    Expression
        Simplified expression

    Example
    -------
    >>> from pharmpy.plugins.nonmem import conf
    >>> conf.parameter_names = ['comment', 'basic']
    >>> from pharmpy.modeling import load_example_model, simplify_expression
    >>> model = load_example_model("pheno")
    >>> simplify_expression(model, "Abs(PTVCL)")
    PTVCL
    >>> conf.parameter_names = ['basic']
    """
    expr = sympify(expr)
    d = dict()
    for p in model.parameters:
        if p.fix:
            s = sympy.Float(p.init)
        elif p.upper < 0:
            s = sympy.Symbol(p.name, real=True, negative=True)
            d[s] = p.symbol
        elif p.upper <= 0:
            s = sympy.Symbol(p.name, real=True, nonpositive=True)
            d[s] = p.symbol
        elif p.lower > 0:
            s = sympy.Symbol(p.name, real=True, positive=True)
            d[s] = p.symbol
        elif p.lower >= 0:
            s = sympy.Symbol(p.name, real=True, nonnegative=True)
            d[s] = p.symbol
        else:
            s = sympy.Symbol(p.name, real=True)
            d[s] = p.symbol
        expr = expr.subs(p.symbol, s)
    # Remaining symbols should all be real
    for s in expr.free_symbols:
        if s.is_real is not True:
            new = sympy.Symbol(s.name, real=True)
            expr = expr.subs(s, new)
            d[new] = s
    simp = sympy.simplify(expr).subs(d)  # Subs symbols back to non-constrained
    return simp



from pathlib import Path

import pandas as pd
import sympy

import pharmpy
from pharmpy import (
    Assignment,
    DataInfo,
    EstimationStep,
    EstimationSteps,
    ModelStatements,
    Parameter,
    Parameters,
    RandomVariable,
    RandomVariables,
)
from pharmpy.modeling import (
    create_joint_distribution,
    set_first_order_absorption,
    set_proportional_error_model,
)
from pharmpy.workflows import default_model_database


def create_start_model(dataset_path, modeltype='pk_oral', cl_init=0.01, vc_init=1.0, mat_init=0.1):
    dataset_path = Path(dataset_path)

    pop_cl = Parameter('POP_CL', cl_init, lower=0)
    pop_vc = Parameter('POP_VC', vc_init, lower=0)
    iiv_cl = Parameter('IIV_CL', 0.01)
    iiv_vc = Parameter('IIV_VC', 0.01)

    params = Parameters([pop_cl, pop_vc, iiv_cl, iiv_vc])

    eta_cl = RandomVariable.normal('eta_cl', 'iiv', 0, iiv_cl.symbol)
    eta_vc = RandomVariable.normal('eta_vc', 'iiv', 0, iiv_vc.symbol)
    rvs = RandomVariables([eta_cl, eta_vc])

    CL = sympy.Symbol('CL')
    VC = sympy.Symbol('VC')
    cl_ass = Assignment(CL, pop_cl.symbol * sympy.exp(eta_cl.symbol))
    vc_ass = Assignment(VC, pop_vc.symbol * sympy.exp(eta_vc.symbol))

    dose = sympy.Symbol('AMT')
    odes = pharmpy.CompartmentalSystem()
    central = odes.add_compartment('CENTRAL')
    output = odes.add_compartment('OUTPUT')
    odes.add_flow(central, output, CL / VC)
    central.dose = pharmpy.Bolus(dose)

    ipred = Assignment('IPRED', central.amount / VC)
    y_ass = Assignment('Y', ipred.symbol)

    stats = ModelStatements([cl_ass, vc_ass, odes, ipred, y_ass])

    datainfo_path = dataset_path.with_suffix('.datainfo')
    separator = r'\s+'
    if datainfo_path.is_file():
        di = DataInfo.read_json(dataset_path.with_suffix('.datainfo'))
        di.separator = separator
    else:
        # FIXME: Create a default di here?
        di = None
    df = pd.read_table(dataset_path, sep=separator, engine='python')

    est = EstimationStep(
        "FOCE",
        interaction=True,
        maximum_evaluations=99999,
        predictions=['CIPREDI'],
        residuals=['CWRES'],
    )
    eststeps = EstimationSteps([est])

    model = pharmpy.Model()
    model.name = 'start'
    model.parameters = params
    model.random_variables = rvs
    model.statements = stats
    model.dependent_variable = y_ass.symbol
    model.database = default_model_database()
    model.dataset = df
    if di:
        model.datainfo = di
    model.estimation_steps = eststeps
    model.filename_extension = '.mod'  # Should this really be needed?

    set_proportional_error_model(model, zero_protection=True)
    create_joint_distribution(model, [eta_cl.name, eta_vc.name])
    if modeltype == 'pk_oral':
        set_first_order_absorption(model)
        model.parameters['POP_MAT'].init = mat_init

    return model



import numpy as np
import pandas as pd
import scipy.linalg
import symengine
import sympy

from pharmpy.statements import sympify

from .expressions import (
    calculate_epsilon_gradient_expression,
    calculate_eta_gradient_expression,
    get_individual_prediction_expression,
    get_population_prediction_expression,
)


def evaluate_expression(model, expression):
    """Evaluate expression using model

    Calculate the value of expression for each data record.
    The expression can contain dataset columns, variables in model and
    population parameters. If the model has parameter estimates these
    will be used. Initial estimates will be used for non-estimated parameters.

    Parameters
    ----------
    model : Model
        Pharmpy model
    expression : str or sympy expression
        Expression to evaluate

    Returns
    -------
    pd.Series
        A series of one evaluated value for each data record

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_expression
    >>> model = load_example_model("pheno")
    >>> evaluate_expression(model, "TVCL*1000")
    0      6.573770
    1      6.573770
    2      6.573770
    3      6.573770
    4      6.573770
             ...
    739    5.165105
    740    5.165105
    741    5.165105
    742    5.165105
    743    5.165105
    Length: 744, dtype: float64

    """
    expression = sympify(expression)
    full_expr = model.statements.before_odes.full_expression(expression)
    pe = model.modelfit_results.parameter_estimates
    inits = model.parameters.inits
    expr = full_expr.subs(dict(pe)).subs(inits)
    data = model.dataset
    expr = symengine.sympify(expr)

    def func(row):
        subs = expr.subs(dict(row))
        return np.float64(subs.evalf())

    df = data.apply(func, axis=1)
    return df


def evaluate_population_prediction(model, parameters=None, dataset=None):
    """Evaluate the numeric population prediction

    The prediction is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The evaluation is done for each data record in the model dataset
    or optionally using the dataset argument.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.Series
        Population predictions

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_population_prediction
    >>> model = load_example_model("pheno_linear")
    >>> evaluate_population_prediction(model)
    0      17.529739
    1      28.179910
    2       9.688648
    3      17.798916
    4      25.023225
             ...
    150    22.459036
    151    29.223295
    152    20.217288
    153    28.472888
    154    34.226455
    Name: PRED, Length: 155, dtype: float64

    See also
    --------
    evaluate_individual_prediction : Evaluate the individual prediction
    """
    y = get_population_prediction_expression(model)

    if parameters is not None:
        y = y.subs(parameters)
    else:
        if model.modelfit_results is not None:
            y = y.subs(model.modelfit_results.parameter_estimates.to_dict())
        else:
            y = y.subs(model.parameters.inits)

    if dataset is not None:
        df = dataset
    else:
        df = model.dataset

    pred = df.apply(lambda row: np.float64(y.subs(row.to_dict())), axis=1)
    pred.name = 'PRED'
    return pred


def evaluate_individual_prediction(model, etas=None, parameters=None, dataset=None):
    """Evaluate the numeric individual prediction

    The prediction is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The evaluation is done for each data record in the model dataset
    or optionally using the dataset argument.
    The evaluation is done at the current eta values
    or optionally at the given eta values.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    etas : dict
        Optional dictionary of eta values
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.Series
        Individual predictions

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_individual_prediction
    >>> model = load_example_model("pheno_linear")
    >>> evaluate_individual_prediction(model)
    0      17.771084
    1      28.881859
    2      11.441728
    3      21.113050
    4      29.783055
             ...
    150    25.375041
    151    31.833395
    152    22.876707
    153    31.905095
    154    38.099690
    Name: IPRED, Length: 155, dtype: float64

    See also
    --------
    evaluate_population_prediction : Evaluate the population prediction
    """

    y = get_individual_prediction_expression(model)
    if parameters is not None:
        y = y.subs(parameters)
    else:
        y = y.subs(model.parameters.inits)

    if dataset is not None:
        df = dataset
    else:
        df = model.dataset

    idcol = model.datainfo.id_column.name

    if etas is None:
        if (
            model.modelfit_results is not None
            and model.modelfit_results.individual_estimates is not None
        ):
            etas = model.modelfit_results.individual_estimates
        elif model.initial_individual_estimates is not None:
            etas = model.initial_individual_estimates
        else:
            etas = pd.DataFrame(
                0,
                index=df[idcol].unique(),
                columns=[eta.name for eta in model.random_variables.etas],
            )

    def fn(row):
        row = row.to_dict()
        curetas = etas.loc[row[idcol]].to_dict()
        a = np.float64(y.subs(row).subs(curetas))
        return a

    ipred = df.apply(fn, axis=1)
    ipred.name = 'IPRED'
    return ipred


def _replace_parameters(model, y, parameters):
    if parameters is not None:
        y = [x.subs(parameters) for x in y]
    else:
        y = [x.subs(model.parameters.inits) for x in y]
    return y


def evaluate_eta_gradient(model, etas=None, parameters=None, dataset=None):
    """Evaluate the numeric eta gradient

    The gradient is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The gradient is done for each data record in the model dataset
    or optionally using the dataset argument.
    The gradient is done at the current eta values
    or optionally at the given eta values.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    etas : dict
        Optional dictionary of eta values
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.DataFrame
        Gradient

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_eta_gradient
    >>> model = load_example_model("pheno_linear")
    >>> evaluate_eta_gradient(model)
         dF/dETA(1)  dF/dETA(2)
    0     -0.159537  -17.609116
    1     -9.325893  -19.562289
    2     -0.104417  -11.346161
    3     -4.452951  -16.682310
    4    -10.838840  -18.981836
    ..          ...         ...
    150   -5.424423  -19.973013
    151  -14.497185  -17.344797
    152   -0.198714  -22.697161
    153   -7.987731  -23.941806
    154  -15.817067  -22.309945
    <BLANKLINE>
    [155 rows x 2 columns]

    See also
    --------
    evaluate_epsilon_gradient : Evaluate the epsilon gradient
    """

    y = calculate_eta_gradient_expression(model)
    y = _replace_parameters(model, y, parameters)

    if dataset is not None:
        df = dataset
    else:
        df = model.dataset
    idcol = model.datainfo.id_column.name

    if etas is None:
        if (
            model.modelfit_results is not None
            and model.modelfit_results.individual_estimates is not None
        ):
            etas = model.modelfit_results.individual_estimates
        elif model.initial_individual_estimates is not None:
            etas = model.initial_individual_estimates
        else:
            etas = pd.DataFrame(
                0,
                index=df[idcol].unique(),
                columns=[eta.name for eta in model.random_variables.etas],
            )

    def fn(row):
        row = row.to_dict()
        curetas = etas.loc[row[idcol]].to_dict()
        a = [np.float64(x.subs(row).subs(curetas)) for x in y]
        return a

    derivative_names = [f'dF/d{eta.name}' for eta in model.random_variables.etas]
    grad = df.apply(fn, axis=1, result_type='expand')
    grad = pd.DataFrame(grad)
    grad.columns = derivative_names
    return grad


def evaluate_epsilon_gradient(model, etas=None, parameters=None, dataset=None):
    """Evaluate the numeric epsilon gradient

    The gradient is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The gradient is done for each data record in the model dataset
    or optionally using the dataset argument.
    The gradient is done at the current eta values
    or optionally at the given eta values.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    etas : dict
        Optional dictionary of eta values
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.DataFrame
        Gradient

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_epsilon_gradient
    >>> model = load_example_model("pheno_linear")
    >>> evaluate_epsilon_gradient(model)
         dY/dEPS(1)
    0     17.771084
    1     28.881859
    2     11.441728
    3     21.113050
    4     29.783055
    ..          ...
    150   25.375041
    151   31.833395
    152   22.876707
    153   31.905095
    154   38.099690
    <BLANKLINE>
    [155 rows x 1 columns]

    See also
    --------
    evaluate_eta_gradient : Evaluate the eta gradient
    """

    y = calculate_epsilon_gradient_expression(model)
    y = _replace_parameters(model, y, parameters)
    eps_names = [eps.name for eps in model.random_variables.epsilons]
    repl = {sympy.Symbol(eps): 0 for eps in eps_names}
    y = [x.subs(repl) for x in y]

    if dataset is not None:
        df = dataset
    else:
        df = model.dataset

    idcol = model.datainfo.id_column.name

    if etas is None:
        if (
            model.modelfit_results is not None
            and model.modelfit_results.individual_estimates is not None
        ):
            etas = model.modelfit_results.individual_estimates
        elif model.initial_individual_estimates is not None:
            etas = model.initial_individual_estimates
        else:
            etas = pd.DataFrame(
                0,
                index=df[idcol].unique(),
                columns=[eta.name for eta in model.random_variables.etas],
            )

    def fn(row):
        row = row.to_dict()
        curetas = etas.loc[row[idcol]].to_dict()
        a = [np.float64(x.subs(row).subs(curetas)) for x in y]
        return a

    grad = df.apply(fn, axis=1, result_type='expand')
    derivative_names = [f'dY/d{eps}' for eps in eps_names]
    grad = pd.DataFrame(grad)
    grad.columns = derivative_names
    return grad


def evaluate_weighted_residuals(model, parameters=None, dataset=None):
    """Evaluate the weighted residuals

    The residuals is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The residuals is done for each data record in the model dataset
    or optionally using the dataset argument.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.Series
        WRES

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_weighted_residuals
    >>> model = load_example_model("pheno_linear")
    >>> evaluate_weighted_residuals(model)
    0     -0.313859
    1      0.675721
    2     -1.544240
    3      1.921720
    4      1.517677
            ...
    150    1.223935
    151   -0.053334
    152   -0.007023
    153    0.931252
    154    0.778389
    Name: WRES, Length: 155, dtype: float64
    """

    omega = model.random_variables.etas.covariance_matrix
    sigma = model.random_variables.epsilons.covariance_matrix
    if parameters is None:
        if model.modelfit_results is not None:
            parameters = model.modelfit_results.parameter_estimates.to_dict()
        else:
            parameters = model.parameters.inits
    omega = omega.subs(parameters)
    sigma = sigma.subs(parameters)
    omega = np.float64(omega)
    sigma = np.float64(sigma)
    if dataset is not None:
        df = dataset
    else:
        df = model.dataset
    # FIXME: Could have option to gradients to set all etas 0
    etas = pd.DataFrame(
        0,
        index=df[model.datainfo.id_column.name].unique(),
        columns=[eta.name for eta in model.random_variables.etas],
    )
    G = evaluate_eta_gradient(model, etas=etas, parameters=parameters, dataset=dataset)
    H = evaluate_epsilon_gradient(model, etas=etas, parameters=parameters, dataset=dataset)
    F = evaluate_population_prediction(model)
    index = df[model.datainfo.id_column.name]
    G.index = index
    H.index = index
    F.index = index
    WRES = np.float64([])
    for i in df[model.datainfo.id_column.name].unique():
        Gi = np.float64(G.loc[[i]])
        Hi = np.float64(H.loc[[i]])
        Fi = F.loc[i:i]
        DVi = np.float64(df['DV'][df[model.datainfo.id_column.name] == i])
        Ci = Gi @ omega @ Gi.T + np.diag(np.diag(Hi @ sigma @ Hi.T))
        WRESi = scipy.linalg.sqrtm(scipy.linalg.inv(Ci)) @ (DVi - Fi)
        WRES = np.concatenate((WRES, WRESi))
    ser = pd.Series(WRES)
    ser.name = 'WRES'
    return ser



"""
:meta private:
"""

import warnings

from pharmpy.modeling import remove_unused_parameters_and_rvs
from pharmpy.modeling.help_functions import _format_input_list
from pharmpy.random_variables import RandomVariables


def remove_iov(model, to_remove=None):
    """Removes all IOV etas given a list with eta names.

    Parameters
    ----------
    model : Model
        Pharmpy model to remove IOV from.
    to_remove : str, list
        Name/names of IOV etas to remove, e.g. 'ETA_IOV_11'.
        If None, all etas that are IOVs will be removed. None is default.
    Return
    ------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> remove_iov(model)       # doctest: +ELLIPSIS
    <...>

    See also
    --------
    add_iiv
    add_iov
    remove_iiv
    add_pk_iiv

    """
    rvs, sset = model.random_variables, model.statements
    etas = _get_etas(rvs, to_remove)
    if not etas:
        warnings.warn('No IOVs present')
        return model

    for eta in etas:
        sset.subs({eta.symbol: 0})
        del rvs[eta]

    model.random_variables = rvs
    model.statements = sset

    model.modelfit_results = None
    remove_unused_parameters_and_rvs(model)
    model.update_source()
    return model


def _get_etas(rvs, list_of_etas):
    list_of_etas = _format_input_list(list_of_etas)
    etas_all = [eta for eta in rvs if eta.level == 'IOV']
    etas = []
    if list_of_etas is None:
        etas = etas_all
    else:
        for eta_str in list_of_etas:
            eta_remove = rvs[eta_str.upper()]
            if eta_remove not in etas:
                etas.append(eta_remove)
                for eta in etas_all:
                    if eta.parameter_names[0] == eta_remove.parameter_names[0] and eta not in etas:
                        etas.append(eta)

    return RandomVariables(etas)



from pathlib import Path

import pandas as pd

import pharmpy.results
from pharmpy.modeling import (
    add_estimation_step,
    copy_model,
    remove_estimation_step,
    set_ode_solver,
    summarize_modelfit_results,
    update_inits,
)
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow


def create_workflow(methods=None, solvers=None, model=None):
    wf = Workflow()
    wf.name = "estmethod"

    if model is not None:
        start_task = Task('start_estmethod', start, model)
    else:
        start_task = Task('start_estmethod', start)

    task_base_model = Task('create_base_model', _create_base_model)
    wf.add_task(task_base_model, predecessors=start_task)

    wf_fit = create_fit_workflow(n=1)
    wf.insert_workflow(wf_fit, predecessors=task_base_model)

    task_base_model_fit = wf.output_tasks

    if not methods:
        methods = ['foce', 'fo', 'imp', 'impmap', 'its', 'saem', 'laplace', 'bayes']
    elif isinstance(methods, str):
        methods = [methods]
    elif isinstance(methods, list):
        methods = [method.lower() for method in methods]

    if solvers == 'all':
        solvers = [None, 'cvodes', 'dgear', 'dverk', 'ida', 'lsoda', 'lsodi']
    elif isinstance(solvers, str) or not solvers:
        solvers = [solvers]
    elif isinstance(solvers, list):
        solvers = [solver.lower() for solver in solvers]
    if None not in solvers:
        solvers.insert(0, None)

    for method in methods:
        for solver in solvers:
            if solver:
                task_name = f'create_{method.upper()}_{solver.upper()}'
            else:
                task_name = f'create_{method.upper()}'
            if method != 'foce' or solver is not None:
                task_no_update = Task(
                    f'{task_name}_raw_inits', _create_est_model, method, solver, False
                )
                wf.add_task(task_no_update, predecessors=task_base_model_fit)

            task_update = Task(f'{task_name}_update_inits', _create_est_model, method, solver, True)
            wf.add_task(task_update, predecessors=task_base_model_fit)

    wf_fit = create_fit_workflow(n=len(wf.output_tasks))
    wf.insert_workflow(wf_fit, predecessors=wf.output_tasks)

    task_post_process = Task('post_process', post_process)
    wf.add_task(
        task_post_process, predecessors=[start_task] + task_base_model_fit + wf.output_tasks
    )

    return wf


def post_process(*models):
    res_models = []
    for model in models:
        if not model.name.startswith('estmethod_'):
            start_model = model
        else:
            res_models.append(model)

    summary = summarize_modelfit_results(res_models)
    settings = summarize_estimation_steps(res_models)

    res = EstMethodResults(
        summary=summary, settings=settings, start_model=start_model, models=res_models
    )

    return res


def start(model):
    return model


def _create_base_model(model):
    base_model = copy_model(model, 'estmethod_FOCE_raw_inits')
    _clear_estimation_steps(base_model)
    est_settings = _create_est_settings('foce')
    eval_settings = _create_eval_settings()
    add_estimation_step(base_model, **est_settings)
    add_estimation_step(base_model, **eval_settings)
    return base_model


def _create_eval_settings(laplace=False):
    eval_settings = {
        'method': 'imp',
        'interaction': True,
        'laplace': laplace,
        'evaluation': True,
        'maximum_evaluations': 9999,
        'isample': 10000,
        'niter': 10,
        'keep_every_nth_iter': 10,
    }
    return eval_settings


def _create_est_settings(method):
    est_settings = dict()
    interaction = True
    laplace = False
    maximum_evaluations = 9999
    auto = True
    keep_every_nth_iter = 10

    if method == 'laplace':
        est_settings['method'] = 'foce'
        laplace = True
    else:
        est_settings['method'] = method

    est_settings['interaction'] = interaction
    est_settings['laplace'] = laplace
    est_settings['maximum_evaluations'] = maximum_evaluations
    est_settings['auto'] = auto
    est_settings['keep_every_nth_iter'] = keep_every_nth_iter

    return est_settings


def _create_est_model(method, solver, update, model):
    if solver:
        model_name = f'estmethod_{method.upper()}_{solver.upper()}'
    else:
        model_name = f'estmethod_{method.upper()}'
    if update:
        model_name += '_update_inits'
    else:
        model_name += '_raw_inits'
    est_model = copy_model(model, model_name)
    _clear_estimation_steps(est_model)
    if update:
        update_inits(est_model)
    est_settings = _create_est_settings(method)
    if method == 'laplace':
        laplace = True
    else:
        laplace = False
    eval_settings = _create_eval_settings(laplace)
    add_estimation_step(est_model, **est_settings)
    add_estimation_step(est_model, **eval_settings)
    if solver:
        set_ode_solver(est_model, solver)
    return est_model


def _clear_estimation_steps(model):
    while len(model.estimation_steps) > 0:
        remove_estimation_step(model, 0)


class EstMethodResults(pharmpy.results.Results):
    rst_path = Path(__file__).parent / 'report.rst'

    def __init__(self, summary=None, settings=None, best_model=None, start_model=None, models=None):
        self.summary = summary
        self.settings = settings
        self.best_model = best_model
        self.start_model = start_model
        self.models = models

    def sorted_by_ofv(self):
        df = self.summary[['ofv', 'runtime_total', 'estimation_runtime']].sort_values(by=['ofv'])
        return df


def summarize_estimation_steps(models):
    dfs = dict()
    for model in models:
        df = model.estimation_steps.to_dataframe()
        df.index = range(1, len(df) + 1)
        dfs[model.name] = df.drop(columns=['tool_options'])

    return pd.concat(dfs.values(), keys=dfs.keys())



# The NONMEM Model class

import re
import shutil
import warnings
from io import StringIO
from pathlib import Path

import pharmpy.data
import pharmpy.model
import pharmpy.plugins.nonmem
import pharmpy.plugins.nonmem.dataset
from pharmpy.data import DatasetError
from pharmpy.datainfo import ColumnInfo, DataInfo
from pharmpy.estimation import EstimationStep, EstimationSteps
from pharmpy.model import ModelSyntaxError
from pharmpy.modeling.write_csv import write_csv
from pharmpy.parameter import Parameter, Parameters
from pharmpy.plugins.nonmem.results import NONMEMChainedModelfitResults
from pharmpy.plugins.nonmem.table import NONMEMTableFile, PhiTable
from pharmpy.random_variables import RandomVariable, RandomVariables
from pharmpy.statements import Assignment, CompartmentalSystem, ODESystem
from pharmpy.symbols import symbol as S
from pharmpy.workflows import NullModelDatabase, default_model_database

from .advan import compartmental_model
from .nmtran_parser import NMTranParser
from .records.factory import create_record
from .update import (
    update_abbr_record,
    update_ccontra,
    update_estimation,
    update_parameters,
    update_random_variables,
    update_statements,
)


def detect_model(src, *args, **kwargs):
    """Check if src represents a NONMEM control stream
    i.e. check if it is a file that contain $PRO
    """
    if not isinstance(src, str):
        return None
    is_control_stream = re.search(r'^\s*\$PRO', src, re.MULTILINE)
    if is_control_stream:
        return Model
    else:
        return None


def convert_model(model):
    """Convert any model into a NONMEM model"""
    if isinstance(model, Model):
        return model
    from pharmpy.modeling import convert_model

    model = convert_model(model, 'generic')
    code = '$PROBLEM\n'
    code += '$INPUT ' + ' '.join(model.datainfo.names) + '\n'
    code += '$DATA file.csv IGNORE=@\n'
    if model.statements.ode_system is None:
        code += '$PRED\n'
    else:
        code += '$PK\n'
    nm_model = pharmpy.model.Model.create_model(StringIO(code))
    nm_model.random_variables = model.random_variables
    nm_model._parameters = model.parameters
    nm_model._old_parameters = Parameters()
    nm_model.statements = model.statements
    if hasattr(model, 'name'):
        nm_model.name = model.name
    if hasattr(model, 'dependent_variable'):
        nm_model.dependent_variable = model.dependent_variable
    nm_model._data_frame = model.dataset
    nm_model._estimation_steps = model.estimation_steps
    nm_model._datainfo = model.datainfo
    nm_model.observation_transformation = model.observation_transformation
    nm_model.update_source()
    try:
        nm_model.database = model.database
    except AttributeError:
        pass
    if model.statements.ode_system:
        nm_model._compartment_map = {
            name: i for i, name in enumerate(model.statements.ode_system.names, start=1)
        }
    return nm_model


class Model(pharmpy.model.Model):
    def __init__(self, code, path=None, **kwargs):
        self.modelfit_results = None
        parser = NMTranParser()
        if path is None:
            self._name = 'run1'
            self.database = NullModelDatabase()
            self.filename_extension = '.ctl'
        else:
            self._name = path.stem
            self.database = default_model_database(path=path.parent)
            self.filename_extension = path.suffix
        self.control_stream = parser.parse(code)
        self._create_datainfo()
        self._initial_individual_estimates_updated = False
        self._updated_etas_file = None
        self._dataset_updated = False
        self._modelfit_results = None
        self._parent_model = None
        self.dependent_variable = S('Y')
        self.observation_transformation = self.dependent_variable
        self._old_observation_transformation = self.dependent_variable

    def __eq__(self, other):
        # Compare NONMEM code to see if models are the same
        self.update_source()
        other.update_source()
        return str(self) == str(other)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        m = re.search(r'.*?(\d+)$', new_name)
        if m:
            n = int(m.group(1))
            for table in self.control_stream.get_records('TABLE'):
                table_path = table.path
                table_name = table_path.stem
                m = re.search(r'(.*?)(\d+)$', table_name)
                if m:
                    table_stem = m.group(1)
                    new_table_name = f'{table_stem}{n}'
                    table.path = table_path.parent / new_table_name
        self._name = new_name

    @property
    def modelfit_results(self):
        if self._modelfit_results is None:
            self.read_modelfit_results()
        return self._modelfit_results

    @modelfit_results.setter
    def modelfit_results(self, res):
        self._modelfit_results = res

    def update_source(self, path=None, force=False, nofiles=False):
        """Update the source

        path - path to modelfile
        nofiles - Set to not write any files (i.e. dataset, phi input etc)
        """
        self._update_initial_individual_estimates(path, nofiles)
        if hasattr(self, '_random_variables'):
            # FIXME: better solution would be to have system for handling dummy parameters etc.
            if not self.random_variables.etas:
                omega = Parameter('DUMMYOMEGA', init=0, fix=True)
                eta = RandomVariable.normal('eta_dummy', 'iiv', 0, omega.symbol)
                statement = Assignment('DUMMYETA', eta.name)
                self.statements.insert(0, statement)
                self.random_variables.append(eta)
                self.parameters.append(omega)
            update_random_variables(self, self._old_random_variables, self._random_variables)
            self._old_random_variables = self._random_variables.copy()
        if hasattr(self, '_parameters'):
            update_parameters(self, self._old_parameters, self._parameters)
            self._old_parameters = self._parameters.copy()
        trans = self.parameter_translation(reverse=True, remove_idempotent=True, as_symbols=True)
        rv_trans = self.rv_translation(reverse=True, remove_idempotent=True, as_symbols=True)
        trans.update(rv_trans)
        if pharmpy.plugins.nonmem.conf.write_etas_in_abbr:
            abbr_trans = self._abbr_translation(rv_trans)
            trans.update(abbr_trans)
        if trans:
            self.statements  # Read statements unless read
        if hasattr(self, '_statements'):
            update_statements(self, self._old_statements, self._statements, trans)
            self._old_statements = self._statements.copy()

        if self._dataset_updated or self.datainfo != self._old_datainfo:
            # FIXME: If no name set use the model name. Set that when setting dataset to input!
            if self.datainfo.path is None:  # or self.datainfo.path == self._old_datainfo.path:
                if path is not None:
                    dir_path = path.parent
                else:
                    dir_path = self.name + ".csv"
                if not nofiles:
                    datapath = write_csv(self, path=dir_path, force=force)
                    self.datainfo.path = datapath.name
                else:
                    self.datainfo.path = Path(dir_path)
            data_record = self.control_stream.get_records('DATA')[0]

            label = self.datainfo.names[0]
            data_record.ignore_character_from_header(label)
            self._update_input()

            # Remove IGNORE/ACCEPT. Could do diff between old dataset and find simple
            # IGNOREs to add i.e. for filter out certain ID.
            del data_record.ignore
            del data_record.accept
            self._dataset_updated = False
            self._old_datainfo = self.datainfo.copy()

            path = self.datainfo.path
            if path is not None:
                assert (
                    not path.exists() or path.is_file()
                ), f'input path change, but no file exists at target {str(path)}'
                data_record = self.control_stream.get_records('DATA')[0]
                data_record.filename = str(path)

        self._update_sizes()
        update_estimation(self)

        if self.observation_transformation != self._old_observation_transformation:
            if not nofiles:
                update_ccontra(self, path, force)

    def _abbr_translation(self, rv_trans):
        abbr_pharmpy = self.control_stream.abbreviated.translate_to_pharmpy_names()
        abbr_replace = self.control_stream.abbreviated.replace
        abbr_trans = update_abbr_record(self, rv_trans)
        abbr_recs = {
            S(abbr_pharmpy[value]): S(key)
            for key, value in abbr_replace.items()
            if value in abbr_pharmpy.keys()
        }
        abbr_trans.update(abbr_recs)
        return abbr_trans

    def _update_sizes(self):
        """Update $SIZES if needed"""
        all_sizes = self.control_stream.get_records('SIZES')
        if len(all_sizes) == 0:
            sizes = create_record('$SIZES ')
        else:
            sizes = all_sizes[0]
        odes = self.statements.ode_system
        if odes is not None and isinstance(odes, CompartmentalSystem):
            n_compartments = len(odes)
            sizes.PC = n_compartments
        if len(all_sizes) == 0 and len(str(sizes)) > 7:
            self.control_stream.insert_record(str(sizes))

    def _update_initial_individual_estimates(self, path, nofiles=False):
        """Update $ETAS

        Could have 0 FIX in model. Need to read these
        """
        if path is None:  # What to do here?
            phi_path = Path('.')
        else:
            phi_path = path.parent
        phi_path /= f'{self.name}_input.phi'

        if self._initial_individual_estimates_updated:
            etas = self.initial_individual_estimates
            zero_fix = self._zero_fix_rvs(eta=True)
            if zero_fix:
                for eta in zero_fix:
                    etas[eta] = 0
            etas = self._sort_eta_columns(etas)
            if not nofiles:
                phi = PhiTable(df=etas)
                table_file = NONMEMTableFile(tables=[phi])
                table_file.write(phi_path)
            # FIXME: This is a common operation
            eta_records = self.control_stream.get_records('ETAS')
            if eta_records:
                record = eta_records[0]
            else:
                record = self.control_stream.append_record('$ETAS ')
            record.path = phi_path
        elif self._updated_etas_file:
            eta_records = self.control_stream.get_records('ETAS')
            if eta_records:
                record = eta_records[0]
            else:
                record = self.control_stream.append_record('$ETAS')
            shutil.copy(self._updated_etas_file, phi_path)
            record.path = phi_path

        if self._initial_individual_estimates_updated or self._updated_etas_file:
            first_est_record = self.control_stream.get_records('ESTIMATION')[0]
            try:
                first_est_record.option_pairs['MCETA']
            except KeyError:
                first_est_record.set_option('MCETA', 1)
            self._updated_etas_file = None
            self._initial_individual_estimates_updated = False

    def validate(self):
        """Validates NONMEM model (records) syntactically."""
        self.control_stream.validate()

    @property
    def parameters(self):
        """Get the Parameters of all parameters"""
        try:
            return self._parameters
        except AttributeError:
            pass

        self._read_parameters()
        if (
            'comment' in pharmpy.plugins.nonmem.conf.parameter_names
            or 'abbr' in pharmpy.plugins.nonmem.conf.parameter_names
        ):
            self.statements
            # reading statements might change parameters. Resetting _old_parameters
            self._old_parameters = self._parameters.copy()

        if not self.random_variables.validate_parameters(self._parameters.inits):
            nearest = self.random_variables.nearest_valid_parameters(self._parameters.inits)
            before, after = self._compare_before_after_params(self._parameters.inits, nearest)
            warnings.warn(
                f"Adjusting initial estimates to create positive semidefinite "
                f"omega/sigma matrices.\nBefore adjusting:  {before}.\n"
                f"After adjusting: {after}"
            )
            self._parameters.inits = nearest
            self._old_parameters = self._parameters.copy()
        return self._parameters

    def _read_parameters(self):
        next_theta = 1
        params = Parameters()
        for theta_record in self.control_stream.get_records('THETA'):
            thetas = theta_record.parameters(next_theta, seen_labels=set(params.names))
            params.extend(thetas)
            next_theta += len(thetas)
        next_omega = 1
        previous_size = None
        for omega_record in self.control_stream.get_records('OMEGA'):
            omegas, next_omega, previous_size = omega_record.parameters(
                next_omega, previous_size, seen_labels=set(params.names)
            )
            params.extend(omegas)
        next_sigma = 1
        previous_size = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            sigmas, next_sigma, previous_size = sigma_record.parameters(
                next_sigma, previous_size, seen_labels=set(params.names)
            )
            params.extend(sigmas)
        self._parameters = params
        self._old_parameters = params.copy()

    @parameters.setter
    def parameters(self, params):
        """params can be a Parameters or a dict-like with name: value

        Current restrictions:
         * Parameters only supported for new initial estimates and fix
         * Only set the exact same parameters. No additions and no removing of parameters
        """
        self.parameters
        super(Model, self.__class__).parameters.fset(self, params)

    @property
    def initial_individual_estimates(self):
        """Initial individual estimates

        These are taken from the $ETAS FILE. 0 FIX ETAs are removed.
        If no $ETAS is present None will be returned.

        Setter assumes that all IDs are present
        """
        try:
            return self._initial_individual_estimates
        except AttributeError:
            pass
        etas = self.control_stream.get_records('ETAS')
        if etas:
            path = Path(etas[0].path)
            if not path.is_absolute():
                source_dir = self.database.retrieve_file(
                    self.name, self.name + self.filename_extension
                ).parent
                path = source_dir / path
                path = path.resolve()
            phi_tables = NONMEMTableFile(path)
            rv_names = [rv.name for rv in self.random_variables if rv.name.startswith('ETA')]
            etas = next(phi_tables).etas[rv_names]
            self._initial_individual_estimates = etas
        else:
            self._initial_individual_estimates = None
        return self._initial_individual_estimates

    @initial_individual_estimates.setter
    def initial_individual_estimates(self, estimates):
        rv_names = {rv.name for rv in self.random_variables if rv.name.startswith('ETA')}
        columns = set(estimates.columns)
        if columns < rv_names:
            raise ValueError(
                f'Cannot set initial estimate for random variable not in the model:'
                f' {rv_names - columns}'
            )
        diff = columns - rv_names
        # If not setting all etas automatically set remaining to 0 for all individuals
        if len(diff) > 0:
            for name in diff:
                estimates = estimates.copy(deep=True)
                estimates[name] = 0
            estimates = self._sort_eta_columns(estimates)
        self._initial_individual_estimates = estimates
        self._initial_individual_estimates_updated = True
        self._updated_etas_file = None

    def _sort_eta_columns(self, df):
        return df.reindex(sorted(df.columns), axis=1)

    @property
    def statements(self):
        try:
            return self._statements
        except AttributeError:
            pass

        rec = self.get_pred_pk_record()
        statements = rec.statements

        des = self._get_des_record()
        error = self._get_error_record()
        if error:
            sub = self.control_stream.get_records('SUBROUTINES')[0]
            advan = sub.get_option_startswith('ADVAN')
            trans = sub.get_option_startswith('TRANS')
            if not trans:
                trans = 'TRANS1'
            comp = compartmental_model(self, advan, trans, des)
            if comp is not None:
                cm, link = comp
                statements += [cm, link]
            else:
                statements.append(ODESystem())  # FIXME: Placeholder for ODE-system
                # FIXME: Dummy link statement
                statements.append(Assignment('F', S('F')))
            statements += error.statements

        if not hasattr(self, '_parameters'):
            self._read_parameters()

        trans_statements, trans_params = self._create_name_trans(statements)
        for key, value in trans_params.items():
            try:
                self.parameters[key].name = value
                for theta in self.control_stream.get_records('THETA'):
                    theta.update_name_map(trans_params)
                for omega in self.control_stream.get_records('OMEGA'):
                    omega.update_name_map(trans_params)
                for sigma in self.control_stream.get_records('SIGMA'):
                    sigma.update_name_map(trans_params)
            except KeyError:
                self.random_variables.subs({S(key): value})

        statements.subs(trans_statements)

        self._statements = statements
        self._old_statements = statements.copy()
        return statements

    @statements.setter
    def statements(self, statements_new):
        self.statements  # Read in old statements
        self._statements = statements_new

    def get_pred_pk_record(self):
        pred = self.control_stream.get_records('PRED')

        if not pred:
            pk = self.control_stream.get_records('PK')
            if not pk:
                raise ModelSyntaxError('Model has no $PK or $PRED')
            return pk[0]
        else:
            return pred[0]

    def _get_pk_record(self):
        pk = self.control_stream.get_records('PK')
        if pk:
            pk = pk[0]
        return pk

    def _get_error_record(self):
        error = self.control_stream.get_records('ERROR')
        if error:
            error = error[0]
        return error

    def _get_des_record(self):
        des = self.control_stream.get_records('DES')
        if des:
            des = des[0]
        return des

    def _create_name_trans(self, statements):
        rvs = self.random_variables

        conf_functions = {
            'comment': self._name_as_comments(statements),
            'abbr': self._name_as_abbr(rvs),
            'basic': self._name_as_basic(),
        }

        abbr = self.control_stream.abbreviated.replace
        pset_current = {
            **self.parameter_translation(reverse=True),
            **{rv: rv for rv in [rv.name for rv in rvs]},
        }
        sset_current = {
            **abbr,
            **{
                rv.name: rv.name
                for rv in rvs
                if rv.name not in abbr.keys() and rv.symbol in statements.free_symbols
            },
            **{
                p: p
                for p in pset_current.values()
                if p not in abbr.keys() and S(p) in statements.free_symbols
            },
        }

        trans_sset, trans_pset = dict(), dict()
        names_sset_translated, names_pset_translated, names_basic = [], [], []
        clashing_symbols = set()

        for setting in pharmpy.plugins.nonmem.conf.parameter_names:
            trans_sset_setting, trans_pset_setting = conf_functions[setting]
            if setting != 'basic':
                clashing_symbols.update(
                    self._clashing_symbols(statements, {**trans_sset_setting, **trans_pset_setting})
                )
            for name_current, name_new in trans_sset_setting.items():
                name_nonmem = sset_current[name_current]

                if S(name_new) in clashing_symbols or name_nonmem in names_sset_translated:
                    continue

                name_in_sset_current = {v: k for k, v in sset_current.items()}[name_nonmem]
                trans_sset[name_in_sset_current] = name_new
                names_sset_translated.append(name_nonmem)

                if name_nonmem in pset_current.values() and name_new in pset_current.keys():
                    names_pset_translated.append(name_nonmem)

            for name_current, name_new in trans_pset_setting.items():
                name_nonmem = pset_current[name_current]

                if S(name_new) in clashing_symbols or name_nonmem in names_pset_translated:
                    continue

                trans_pset[name_current] = name_new
                names_pset_translated.append(name_nonmem)

            if setting == 'basic':
                params_left = [k for k in pset_current.keys() if k not in names_pset_translated]
                params_left += [rv.name for rv in rvs if rv.name not in names_sset_translated]
                names_basic = [name for name in params_left if name not in names_sset_translated]
                break

        if clashing_symbols:
            warnings.warn(
                f'The parameter names {clashing_symbols} are also names of variables '
                f'in the model code. Falling back to the in naming scheme config '
                f'names for these.'
            )

        names_nonmem_all = [rv.name for rv in rvs] + [
            key for key in self.parameter_translation().keys()
        ]

        if set(names_nonmem_all) - set(names_sset_translated + names_pset_translated + names_basic):
            raise ValueError(
                'Mismatch in number of parameter names, all have not been accounted for. If basic '
                'NONMEM-names are desired as fallback, double-check that "basic" is included in '
                'config-settings for parameter_names.'
            )
        return trans_sset, trans_pset

    def _name_as_comments(self, statements):
        params_current = self.parameter_translation(remove_idempotent=True)
        for name_abbr, name_nonmem in self.control_stream.abbreviated.replace.items():
            if name_nonmem in params_current.keys():
                params_current[name_abbr] = params_current.pop(name_nonmem)
        trans_params = {
            name_comment: name_comment
            for name_current, name_comment in params_current.items()
            if S(name_current) not in statements.free_symbols
        }
        trans_statements = {
            name_current: name_comment
            for name_current, name_comment in params_current.items()
            if S(name_current) in statements.free_symbols
        }
        return trans_statements, trans_params

    def _name_as_abbr(self, rvs):
        pharmpy_names = self.control_stream.abbreviated.translate_to_pharmpy_names()
        params_current = self.parameter_translation(remove_idempotent=True, reverse=True)
        trans_params = {
            name_nonmem: name_abbr
            for name_nonmem, name_abbr in pharmpy_names.items()
            if name_nonmem in self.parameter_translation().keys()
            or name_nonmem in [rv.name for rv in rvs]
        }
        for name_nonmem, name_abbr in params_current.items():
            if name_abbr in trans_params.keys():
                trans_params[name_nonmem] = trans_params.pop(name_abbr)
        trans_statements = {
            name_abbr: pharmpy_names[name_nonmem]
            for name_abbr, name_nonmem in self.control_stream.abbreviated.replace.items()
        }
        return trans_statements, trans_params

    def _name_as_basic(self):
        trans_params = {
            name_current: name_nonmem
            for name_current, name_nonmem in self.parameter_translation(reverse=True).items()
            if name_current != name_nonmem
        }
        trans_statements = self.control_stream.abbreviated.replace
        return trans_statements, trans_params

    @staticmethod
    def _clashing_symbols(statements, trans_statements):
        parameter_symbols = {S(symb) for _, symb in trans_statements.items()}
        clashing_symbols = parameter_symbols & statements.free_symbols
        return clashing_symbols

    def replace_abbr(self, replace):
        for key, value in replace.items():
            try:
                self.parameters[key].name = value
            except KeyError:
                pass
        self.random_variables.rename(replace)

    def _zero_fix_rvs(self, eta=True):
        zero_fix = []
        if eta:
            prev_cov = None
            next_omega = 1
            for omega_record in self.control_stream.get_records('OMEGA'):
                _, next_omega, prev_cov, new_zero_fix = omega_record.random_variables(
                    next_omega, prev_cov
                )
                zero_fix += new_zero_fix
        else:
            prev_cov = None
            next_sigma = 1
            for sigma_record in self.control_stream.get_records('SIGMA'):
                _, next_sigma, prev_cov, new_zero_fix = sigma_record.random_variables(
                    next_sigma, prev_cov
                )
                zero_fix += new_zero_fix
        return zero_fix

    @property
    def random_variables(self):
        try:
            return self._random_variables
        except AttributeError:
            pass
        rvs = RandomVariables()
        next_omega = 1
        prev_cov = None

        for omega_record in self.control_stream.get_records('OMEGA'):
            etas, next_omega, prev_cov, _ = omega_record.random_variables(next_omega, prev_cov)
            rvs.extend(etas)
        self.adjust_iovs(rvs)
        next_sigma = 1
        prev_cov = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            epsilons, next_sigma, prev_cov, _ = sigma_record.random_variables(next_sigma, prev_cov)
            rvs.extend(epsilons)
        self._random_variables = rvs
        self._old_random_variables = rvs.copy()

        if (
            'comment' in pharmpy.plugins.nonmem.conf.parameter_names
            or 'abbr' in pharmpy.plugins.nonmem.conf.parameter_names
        ):
            self.statements

        return self._random_variables

    @staticmethod
    def adjust_iovs(rvs):
        for i, rv in enumerate(rvs):
            try:
                next_rv = rvs[i + 1]
            except IndexError:
                break

            if rv.level != 'IOV' and next_rv.level == 'IOV':
                associated_rvs = rv.joint_names
                if len(associated_rvs) > 1:
                    for rv_con in associated_rvs:
                        rvs[rv_con].level = 'IOV'
                else:
                    rv.level = 'IOV'

    @random_variables.setter
    def random_variables(self, new):
        self.random_variables  # Read in old random variables
        self._random_variables = new.copy()

    @property
    def estimation_steps(self):
        try:
            return self._estimation_steps
        except AttributeError:
            pass

        steps = []
        records = self.control_stream.get_records('ESTIMATION')
        covrec = self.control_stream.get_records('COVARIANCE')
        for record in records:
            value = record.get_option('METHOD')
            if value is None or value == '0' or value == 'ZERO':
                name = 'fo'
            elif value == '1' or value == 'CONDITIONAL' or value == 'COND':
                name = 'foce'
            else:
                name = value
            interaction = False
            evaluation = False
            maximum_evaluations = None
            cov = False
            laplace = False
            isample = None
            niter = None
            auto = None
            keep_every_nth_iter = None

            if record.has_option('INTERACTION') or record.has_option('INTER'):
                interaction = True
            maxeval_opt = (
                record.get_option('MAXEVAL') if not None else record.get_option('MAXEVALS')
            )
            if maxeval_opt is not None:
                if (name.upper() == 'FO' or name.upper() == 'FOCE') and int(maxeval_opt) == 0:
                    evaluation = True
                else:
                    maximum_evaluations = int(maxeval_opt)
            eval_opt = record.get_option('EONLY')
            if eval_opt is not None and int(eval_opt) == 1:
                evaluation = True
            if covrec:
                cov = True
            if record.has_option('LAPLACIAN') or record.has_option('LAPLACE'):
                laplace = True
            if record.has_option('ISAMPLE'):
                isample = int(record.get_option('ISAMPLE'))
            if record.has_option('NITER'):
                niter = int(record.get_option('NITER'))
            if record.has_option('AUTO'):
                auto_opt = record.get_option('AUTO')
                if auto_opt is not None and int(auto_opt) in [0, 1]:
                    auto = bool(auto_opt)
                else:
                    raise ValueError('Currently only AUTO=0 and AUTO=1 is supported')
            if record.has_option('PRINT'):
                keep_every_nth_iter = int(record.get_option('PRINT'))

            protected_names = [
                name.upper(),
                'EONLY',
                'INTERACTION',
                'INTER',
                'LAPLACE',
                'LAPLACIAN',
                'MAXEVAL',
                'MAXEVALS',
                'METHOD',
                'METH',
                'ISAMPLE',
                'NITER',
                'AUTO',
                'PRINT',
            ]

            tool_options = {
                option.key: option.value
                for option in record.all_options
                if option.key not in protected_names
            }
            if not tool_options:
                tool_options = None

            try:
                meth = EstimationStep(
                    name,
                    interaction=interaction,
                    cov=cov,
                    evaluation=evaluation,
                    maximum_evaluations=maximum_evaluations,
                    laplace=laplace,
                    isample=isample,
                    niter=niter,
                    auto=auto,
                    keep_every_nth_iter=keep_every_nth_iter,
                    tool_options=tool_options,
                )
            except ValueError:
                raise ModelSyntaxError(f'Non-recognized estimation method in: {str(record.root)}')
            steps.append(meth)

        steps = EstimationSteps(steps)

        # Read eta and epsilon derivatives
        table_records = self.control_stream.get_records('TABLE')
        for table in table_records:
            etaderivs = table.eta_derivatives
            if etaderivs:
                etas = self.random_variables.etas
                etaderiv_names = [etas[i - 1].name for i in etaderivs]
                for step in steps:
                    step.eta_derivatives = etaderiv_names
            epsderivs = table.epsilon_derivatives
            if epsderivs:
                epsilons = self.random_variables.epsilons
                epsilonderivs_names = [epsilons[i - 1].name for i in epsderivs]
                for step in steps:
                    step.epsilon_derivatives = epsilonderivs_names

        self._estimation_steps = steps
        self._old_estimation_steps = steps.copy()
        return steps

    @estimation_steps.setter
    def estimation_steps(self, value):
        self._old_estimation_steps = self._estimation_steps.copy()
        self._estimation_steps = value

    @property
    def model_code(self):
        self.update_source(nofiles=True)
        return str(self.control_stream)

    def _read_dataset_path(self):
        datas = self.control_stream.get_records('DATA')
        if datas:
            record = datas[0]
        else:
            return None
        path = Path(record.filename)
        if not path.is_absolute():
            try:
                dbpath = self.database.retrieve_file(self.name, path)
            except FileNotFoundError:
                pass
            else:
                if dbpath is not None:
                    path = dbpath
        try:
            return path.resolve()
        except FileNotFoundError:
            return path

    @property
    def dataset(self):
        try:
            return self._data_frame
        except AttributeError:
            self._data_frame = self._read_dataset(raw=False)
        return self._data_frame

    @dataset.setter
    def dataset(self, df):
        self._dataset_updated = True
        self._data_frame = df
        self.update_datainfo()

    def read_raw_dataset(self, parse_columns=tuple()):
        return self._read_dataset(raw=True, parse_columns=parse_columns)

    @staticmethod
    def _synonym(key, value):
        """Return a tuple reserved name and synonym"""
        _reserved_column_names = [
            'ID',
            'L1',
            'L2',
            'DV',
            'MDV',
            'RAW_',
            'MRG_',
            'RPT_',
            'TIME',
            'DATE',
            'DAT1',
            'DAT2',
            'DAT3',
            'EVID',
            'AMT',
            'RATE',
            'SS',
            'II',
            'ADDL',
            'CMT',
            'PCMT',
            'CALL',
            'CONT',
        ]
        if key in _reserved_column_names:
            return (key, value)
        elif value in _reserved_column_names:
            return (value, key)
        else:
            raise DatasetError(
                f'A column name "{key}" in $INPUT has a synonym to a non-reserved '
                f'column name "{value}"'
            )

    def _column_info(self):
        """List all column names in order.
        Use the synonym when synonym exists.
        return tuple of two lists, colnames, and drop together with a dictionary
        of replacements for reserved names (aka synonyms).
        Anonymous columns, i.e. DROP or SKIP alone, will be given unique names _DROP1, ...
        """
        input_records = self.control_stream.get_records("INPUT")
        colnames = []
        drop = []
        synonym_replacement = {}
        next_anonymous = 1
        for record in input_records:
            for key, value in record.all_options:
                if value:
                    if key == 'DROP' or key == 'SKIP':
                        drop.append(True)
                        colnames.append(value)
                        reserved_name = value
                    elif value == 'DROP' or value == 'SKIP':
                        colnames.append(key)
                        drop.append(True)
                        reserved_name = key
                    else:
                        drop.append(False)
                        (reserved_name, synonym) = Model._synonym(key, value)
                        synonym_replacement[reserved_name] = synonym
                        colnames.append(synonym)
                else:
                    if key == 'DROP' or key == 'SKIP':
                        drop.append(True)
                        name = f'_DROP{next_anonymous}'
                        colnames.append(name)
                        reserved_name = name
                        next_anonymous += 1
                    else:
                        drop.append(False)
                        colnames.append(key)
                        reserved_name = key
        return colnames, drop, synonym_replacement

    def _update_input(self):
        """Update $INPUT

        currently supporting append columns at end and removing columns
        And add/remove DROP
        """
        new_names = self.datainfo.names
        colnames, drop, _ = self._column_info()
        removed_columns = set(colnames) - set(new_names)
        input_records = self.control_stream.get_records("INPUT")
        for col in removed_columns:
            input_records[0].remove_option(col)
        appended_names = new_names[len(colnames) :]
        last_input_record = input_records[-1]
        for colname in appended_names:
            last_input_record.append_option(colname)
        # Update DROP
        colnames, drop, _ = self._column_info()
        keep = []
        i = 0
        for child in input_records[0].root.children:
            if child.rule == 'option':
                if drop[i] != self.datainfo[colnames[i]].drop:
                    new = input_records[0]._create_option(colnames[i], 'DROP')
                    keep.append(new)
                else:
                    keep.append(child)
                i += 1
            else:
                keep.append(child)
        input_records[0].root.children = keep

    def _replace_synonym_in_filters(filters, replacements):
        result = []
        for f in filters:
            if f.COLUMN in replacements:
                s = ''
                for child in f.children:
                    if child.rule == 'COLUMN':
                        value = replacements[f.COLUMN]
                    else:
                        value = str(child)
                    s += value
            else:
                s = str(f)
            result.append(s)
        return result

    def _create_datainfo(self):
        dataset_path = self._read_dataset_path()
        try:
            path = dataset_path.with_suffix('.datainfo')
        except:  # noqa: E722
            # FIXME: dataset_path could fail in so many ways!
            pass
        else:
            if path.is_file():
                di = DataInfo.read_json(path)
                di.path = dataset_path
                self.datainfo = di
                self._old_datainfo = di.copy()
                return
        (colnames, drop, replacements) = self._column_info()
        column_info = []
        for colname, coldrop in zip(colnames, drop):
            info = ColumnInfo(colname, drop=coldrop)
            if colname == 'ID' or colname == 'L1':
                info.type = 'id'
                info.scale = 'nominal'
                info.datatype = 'int64'
            elif colname == 'DV' or colname == replacements.get('DV', None):
                info.type = 'dv'
            elif colname == 'TIME' or colname == replacements.get('TIME', None):
                info.type = 'idv'
                info.scale = 'ratio'
                info.datatype = 'nmtran-time'
            elif colname in ['DATE', 'DAT1', 'DAT2', 'DAT3']:
                info.scale = 'interval'
                info.datatype = 'nmtran-date'
            elif colname == 'EVID' and self._get_pk_record():
                info.type = 'event'
                info.scale = 'nominal'
            elif colname == 'MDV' and self._get_pk_record():
                if 'EVID' in colnames:
                    info.type = 'mdv'
                else:
                    info.type = 'event'
                info.scale = 'nominal'
            elif colname == 'II' and self._get_pk_record():
                info.type = 'ii'
                info.scale = 'ratio'
            elif colname == 'ADDL' and self._get_pk_record():
                info.type = 'additional'
                info.scale = 'ordinal'
            elif (
                colname == 'AMT' or colname == replacements.get('AMT', None)
            ) and self._get_pk_record():
                info.type = 'dose'
                info.scale = 'ratio'
            column_info.append(info)

        di = DataInfo(column_info)
        di.path = dataset_path
        self.datainfo = di
        self._old_datainfo = di.copy()

    def _read_dataset(self, raw=False, parse_columns=tuple()):
        data_records = self.control_stream.get_records('DATA')
        ignore_character = data_records[0].ignore_character
        null_value = data_records[0].null_value
        (colnames, drop, replacements) = self._column_info()

        if raw:
            ignore = None
            accept = None
        else:
            # FIXME: All direct handling of control stream spanning
            # over one or more records should move
            ignore = data_records[0].ignore
            accept = data_records[0].accept
            # FIXME: This should really only be done if setting the dataset
            if ignore:
                ignore = Model._replace_synonym_in_filters(ignore, replacements)
            else:
                accept = Model._replace_synonym_in_filters(accept, replacements)

        df = pharmpy.plugins.nonmem.dataset.read_nonmem_dataset(
            self.datainfo.path,
            raw,
            ignore_character,
            colnames,
            drop,
            null_value=null_value,
            parse_columns=parse_columns,
            ignore=ignore,
            accept=accept,
        )
        # Let TIME be the idv in both $PK and $PRED models

        # Remove individuals without observations
        try:
            from pharmpy.modeling.data import get_observations

            # This is a hack to be able to use the get_observations function
            # before the dataset has been properly read in.
            self._data_frame = df
            have_obs = set(get_observations(self).index.unique(level=0))
        except DatasetError:
            pass
        else:
            all_ids = set(df['ID'].unique())
            ids_to_remove = all_ids - have_obs
            df = df[~df['ID'].isin(ids_to_remove)]
        return df

    def rv_translation(self, reverse=False, remove_idempotent=False, as_symbols=False):
        self.random_variables
        d = dict()
        for record in self.control_stream.get_records('OMEGA'):
            for key, value in record.eta_map.items():
                nonmem_name = f'ETA({value})'
                d[nonmem_name] = key
        for record in self.control_stream.get_records('SIGMA'):
            for key, value in record.eta_map.items():
                nonmem_name = f'EPS({value})'
                d[nonmem_name] = key
        if remove_idempotent:
            d = {key: val for key, val in d.items() if key != val}
        if reverse:
            d = {val: key for key, val in d.items()}
        if as_symbols:
            d = {S(key): S(val) for key, val in d.items()}
        return d

    def parameter_translation(self, reverse=False, remove_idempotent=False, as_symbols=False):
        """Get a dict of NONMEM name to Pharmpy parameter name
        i.e. {'THETA(1)': 'TVCL', 'OMEGA(1,1)': 'IVCL'}
        """
        self.parameters
        d = dict()
        for theta_record in self.control_stream.get_records('THETA'):
            for key, value in theta_record.name_map.items():
                nonmem_name = f'THETA({value})'
                d[nonmem_name] = key
        for record in self.control_stream.get_records('OMEGA'):
            for key, value in record.name_map.items():
                nonmem_name = f'OMEGA({value[0]},{value[1]})'
                d[nonmem_name] = key
        for record in self.control_stream.get_records('SIGMA'):
            for key, value in record.name_map.items():
                nonmem_name = f'SIGMA({value[0]},{value[1]})'
                d[nonmem_name] = key
        if remove_idempotent:
            d = {key: val for key, val in d.items() if key != val}
        if reverse:
            d = {val: key for key, val in d.items()}
        if as_symbols:
            d = {S(key): S(val) for key, val in d.items()}
        return d

    def read_modelfit_results(self):
        try:
            ext_path = self.database.retrieve_file(self.name, self.name + '.ext')
        except FileNotFoundError:
            self._modelfit_results = None
            return None
        if ext_path is not None:
            self._modelfit_results = NONMEMChainedModelfitResults(ext_path, model=self)
            return self._modelfit_results



from pathlib import Path

import pytest
import sympy
import sympy.physics.units

from pharmpy.datainfo import ColumnInfo, DataInfo


def test_columninfo_init():
    with pytest.raises(TypeError):
        ColumnInfo(1)


def test_columninfo_type():
    col = ColumnInfo("DUMMY")
    with pytest.raises(TypeError):
        col.type = "notaknowntype"
    col.type = 'id'
    assert col.type == 'id'

    col2 = ColumnInfo("DUMMY", type='dv')
    assert col2.type == 'dv'
    assert col2.continuous


def test_columninfo_descriptor():
    col = ColumnInfo("DUMMY", descriptor="body weight")
    assert col.descriptor == "body weight"
    col2 = ColumnInfo("DUMMY2")
    assert col2.descriptor is None
    with pytest.raises(TypeError):
        col2.descriptor = "notaknowndescriptor"


def test_columninfo_scale():
    col = ColumnInfo("DUMMY")
    with pytest.raises(TypeError):
        col.scale = 'notavalidscale'
    col.scale = 'nominal'
    assert col.scale == 'nominal'
    assert not col.continuous


def test_columninfo_unit():
    col = ColumnInfo("DUMMY")
    col.unit = "nospecialunit"
    assert col.unit == sympy.Symbol("nospecialunit")
    col.unit = "kg"
    assert col.unit == sympy.physics.units.kg


def test_columninfo_continuous():
    col = ColumnInfo("DUMMY", scale="nominal")
    with pytest.raises(ValueError):
        col.continuous = True


def test_columninfo_is_numerical():
    col = ColumnInfo("DUMMY", scale='nominal')
    assert not col.is_numerical()
    col = ColumnInfo("DUMMY", scale='ratio')
    assert col.is_numerical()


def test_columninfo_repr():
    col = ColumnInfo("DUMMY", scale='nominal')
    correct = """              DUMMY
type        unknown
scale       nominal
continuous    False
categories     None
unit              1
drop          False
datatype    float64
descriptor     None"""
    assert repr(col) == correct


def test_init():
    di = DataInfo()
    assert len(di) == 0


def test_eq():
    di1 = DataInfo()
    di2 = DataInfo(columns=["COL1", "COL2"])
    assert di1 != di2
    di3 = DataInfo(columns=["DUMMY1", "DUMMY2"])
    assert di2 != di3


def test_indexing():
    di = DataInfo(columns=["COL1", "COL2"])
    assert di[0].name == 'COL1'
    with pytest.raises(TypeError):
        di[1.0]
    col = ColumnInfo("COL3")
    di[0] = col
    assert di["COL3"].name == "COL3"
    del di[0]
    assert len(di) == 1
    di.insert(1, col)
    assert len(di) == 2


def test_id_column():
    di = DataInfo(['ID', 'TIME', 'DV'])
    with pytest.raises(IndexError):
        di.id_column = 'DUMMY'
    di.id_column = 'ID'
    assert di.id_column.name == 'ID'


def test_dv_label():
    di = DataInfo(['ID', 'TIME', 'DV'])
    with pytest.raises(IndexError):
        di.dv_column = 'DUMMY'
    di.dv_column = 'DV'
    assert di.dv_column.name == 'DV'


def test_idv_label():
    di = DataInfo(['ID', 'TIME', 'DV'])
    with pytest.raises(IndexError):
        di.idv_column = 'DUMMY'
    di.idv_column = 'TIME'
    assert di.idv_column.name == 'TIME'


def test_get_set_column_type():
    di = DataInfo(['ID', 'TIME', 'DV'])
    di['ID'].type = 'id'
    with pytest.raises(IndexError):
        di['DUMMY'].type = 'id'
    with pytest.raises(TypeError):
        di['TIME'].type = 'kzarqj'
    assert di['ID'].type == 'id'


def test_get_column_label():
    di = DataInfo(['ID', 'TIME', 'DV', 'WGT', 'APGR'])
    di['ID'].type = 'id'
    di[['WGT', 'APGR']].types = 'covariate'
    assert di.typeix['id'].names == ['ID']
    assert di.typeix['covariate'].names == ['WGT', 'APGR']


def test_unit():
    di = DataInfo(['ID', 'TIME', 'DV', 'WGT', 'APGR'])
    assert di['ID'].unit == 1


def test_scale():
    col = ColumnInfo('WGT', scale='ratio')
    assert col
    with pytest.raises(TypeError):
        ColumnInfo('DUMMY', scale='dummy')


def test_json():
    col1 = ColumnInfo("ID", type='id', scale='nominal')
    col2 = ColumnInfo("TIME", type='idv', scale='ratio', unit="h")
    di = DataInfo([col1, col2])
    correct = '{"columns": [{"name": "ID", "type": "id", "scale": "nominal", "continuous": false, "categories": null, "unit": "1"}, {"name": "TIME", "type": "idv", "scale": "ratio", "continuous": true, "categories": null, "unit": "hour"}]}'  # noqa: E501
    assert di.to_json() == correct

    newdi = DataInfo.from_json(correct)
    assert newdi == di


def test_path():
    di = DataInfo(["C1", "C2"])
    di.path = "file.datainfo"
    assert di.path == Path("file.datainfo")


def test_types():
    di = DataInfo(['ID', 'TIME', 'DV'])
    di.id_column = 'ID'
    di.dv_column = 'DV'
    di.idv_column = 'TIME'
    assert di.types == ['id', 'idv', 'dv']



import shutil
from pathlib import Path

from ..model_database import LocalModelDirectoryDatabase
from .baseclass import ToolDatabase


class LocalDirectoryToolDatabase(ToolDatabase):
    """ToolDatabase in a local directory

    Parameters
    ----------
    toolname : str
        Name of the tool
    path : str or Path
        Path to directory. Will be created if it does not exist.
    """

    def __init__(self, toolname, path=None):
        if path is None:
            i = 1
            while True:
                name = f'{toolname}_dir{i}'
                path = Path(name)
                if not path.exists():
                    break
                i += 1
        path = Path(path)
        path.mkdir(parents=True)
        self.path = path.resolve()

        modeldb = LocalModelDirectoryDatabase(self.path / 'models')
        self.model_database = modeldb
        super().__init__(toolname)

    def store_local_file(self, source_path):
        if Path(source_path).is_file():
            shutil.copy2(source_path, self.path)

    def store_results(self, res):
        res.to_json(path=self.path / 'results.json')
        res.to_csv(path=self.path / 'results.csv')



import shutil

import pytest

from pharmpy import Model
from pharmpy.modeling import run_tool
from pharmpy.utils import TemporaryDirectoryChanger


@pytest.mark.parametrize(
    'methods, solvers, no_of_models, advan_ref',
    [('foce', None, 2, 'ADVAN1'), ('foce', ['lsoda'], 4, 'ADVAN13'), ('foce', ['GL'], 4, '$MODEL')],
)
def test_estmethod(tmp_path, testdata, methods, solvers, no_of_models, advan_ref):
    with TemporaryDirectoryChanger(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model_start = Model.create_model('pheno_real.mod')
        model_start.datainfo.path = tmp_path / 'pheno.dta'

        res = run_tool('estmethod', methods=methods, solvers=solvers, model=model_start)

        assert len(res.summary) == no_of_models
        assert len(res.models) == no_of_models
        assert advan_ref in res.models[-1].model_code
        rundir = tmp_path / 'estmethod_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'results.html').exists()



from pathlib import Path

import pandas as pd

from pharmpy.results import Results


class ResmodResults(Results):
    """Resmod results class"""

    def __init__(self, models=None):
        self.models = models


def calculate_results(base_model, tvar_models, other_models):
    base_ofv = base_model.modelfit_results.ofv

    model_name = []
    model_dofv = []
    model_params = []
    for model in other_models:
        name = model.name
        dofv = base_ofv - model.modelfit_results.ofv
        if name == 'IIV_on_RUV':
            param = {'omega': round(model.modelfit_results.parameter_estimates["IIV_RUV1"], 6)}
        elif name == 'power':
            param = {'theta': round(model.modelfit_results.parameter_estimates["power1"], 6)}
        else:
            param = {
                'sigma_add': round(model.modelfit_results.parameter_estimates["sigma_add"], 6),
                'sigma_prop': round(model.modelfit_results.parameter_estimates["sigma_prop"], 6),
            }
        model_name.append(name)
        model_dofv.append(dofv)
        model_params.append(param)

    df = pd.DataFrame(
        {
            'model': model_name,
            'dvid': 1,
            'iteration': 1,
            'dofv': model_dofv,
            'parameters': model_params,
        }
    )

    tvar_name = []
    dofv_tvar = []
    theta_tvar = []
    for model in tvar_models:
        name = model.name
        dofv = base_ofv - model.modelfit_results.ofv
        theta = round(model.modelfit_results.parameter_estimates["time_varying"], 6)
        tvar_name.append(name)
        dofv_tvar.append(dofv)
        theta_tvar.append(theta)

    params_tvar = []
    for i in range(1, len(tvar_models) + 1):
        param = {f"theta_tvar{i}": theta_tvar[i - 1]}
        params_tvar.append(param)

    df_tvar = pd.DataFrame(
        {
            'model': tvar_name,
            'dvid': 1,
            'iteration': 1,
            'dofv': dofv_tvar,
            'parameters': params_tvar,
        }
    )

    df_final = pd.concat([df, df_tvar])
    df_final.set_index(['model', 'dvid', 'iteration'], inplace=True)

    res = ResmodResults(models=df_final)
    return res


def psn_resmod_results(path):
    path = Path(path)
    res = ResmodResults()
    respath = path / 'resmod_results.csv'
    if respath.is_file():
        df = pd.read_csv(respath, names=range(40), skiprows=[0], engine='python')
    df[0].fillna(1, inplace=True)
    df[1].fillna(1, inplace=True)
    df.dropna(how='all', axis=1, inplace=True)
    df2 = df[[0, 1, 2, 3]].copy()
    df2 = df2.astype({0: int})
    df2.columns = ['iteration', 'DVID', 'model', 'dOFV']
    df2.set_index(['iteration', 'DVID', 'model'], inplace=True)
    parameters = pd.Series(name='parameters', index=df.index, dtype=object)
    for rowind, row in df.iterrows():
        d = dict()
        for i in range(4, len(row)):
            if row[i] is not None:
                a = row[i].split('=')
                d[a[0]] = float(a[1])
        parameters[rowind] = d
    parameters.index = df2.index
    df2['parameters'] = parameters
    res.models = df2
    return res



import sys

import pytest

from pharmpy.modeling import load_example_model, predict_influential_individuals, predict_outliers


@pytest.mark.skipif(
    sys.version_info >= (3, 10), reason="Skipping tests requiring tflite for Python 3.10"
)
def test_predict_outliers():
    model = load_example_model('pheno')
    res = predict_outliers(model)
    assert len(res) == 59
    assert res['residual'][1] == -0.28144291043281555


@pytest.mark.skipif(
    sys.version_info >= (3, 10), reason="Skipping tests requiring tflite for Python 3.10"
)
def test_predict_influential_individuals():
    model = load_example_model('pheno')
    res = predict_influential_individuals(model)
    assert len(res) == 59
    assert res['dofv'][59] == 0.08806940913200378



"""
:meta private:
"""


def update_inits(model, force_individual_estimates=False, move_est_close_to_bounds=False):
    """Update initial parameter estimate for a model

    Updates initial estimates of population parameters for a model from
    its modelfit_results. If the model has used initial estimates for
    individual estimates these will also be updated. If the new initial estimates
    are out of bounds or NaN this function will raise.

    Parameters
    ----------
    model : Model
        Pharmpy model to update initial estimates
    force_individual_estimates : bool
        Update initial individual estimates even if model din't use them previously.
    move_est_close_to_bounds : bool
        Move estimates that are close to bounds. If correlation >0.99 the correlation will
        be set to 0.9, if variance is <0.001 the variance will be set to 0.01.

    Returns
    -------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, update_inits
    >>> model = load_example_model("pheno")   # This model was previously fitted to its data
    >>> model.parameters.inits  # doctest:+ELLIPSIS
    {'THETA(1)': 0.00469307, 'THETA(2)': 1.00916, 'THETA(3)': 0.1, 'OMEGA(1,1)': 0.0309626...}
    >>> update_inits(model)  # doctest:+ELLIPSIS
    <...>
    >>> model.parameters.inits  # doctest:+ELLIPSIS
    {'THETA(1)': 0.00469555, 'THETA(2)': 0.984258, 'THETA(3)': 0.15892, 'OMEGA(1,1)': 0.0293508...}

    """
    if isinstance(model, list) and len(model) == 1:
        model = model[0]

    res = model.modelfit_results
    if not res:
        raise ValueError(
            'Cannot update initial parameter estimates since no modelfit results are available'
        )

    if move_est_close_to_bounds:
        param_est = _move_est_close_to_bounds(model)
    else:
        param_est = res.parameter_estimates

    try:
        model.parameters = param_est
    except ValueError as e:
        if str(e) == 'Initial estimate cannot be set to NaN':
            raise ValueError('One or more parameter estimates are NaN')
        else:
            raise

    if model.initial_individual_estimates is not None or force_individual_estimates:
        model.initial_individual_estimates = res.individual_estimates

    return model


def _move_est_close_to_bounds(model):
    rvs = model.random_variables
    res = model.modelfit_results
    est = res.parameter_estimates.to_dict()
    sdcorr = rvs.parameters_sdcorr(est)
    newdict = est.copy()
    for rvs, dist in rvs.distributions():
        if len(rvs) > 1:
            sigma_sym = dist.sigma
            for i in range(sigma_sym.rows):
                for j in range(sigma_sym.cols):
                    param_name = sigma_sym[i, j].name
                    if i != j:
                        if sdcorr[param_name] > 0.99:
                            name_i, name_j = sigma_sym[i, i].name, sigma_sym[j, j].name
                            # From correlation to covariance
                            corr_new = 0.9
                            sd_i, sd_j = sdcorr[name_i], sdcorr[name_j]
                            newdict[param_name] = corr_new * sd_i * sd_j
                    else:
                        if est[param_name] < 0.001:
                            newdict[param_name] = 0.01
        else:
            param_name = (dist.std**2).name
            if est[param_name] < 0.001:
                newdict[param_name] = 0.01
    return newdict



import numpy as np

from pharmpy.modeling import calculate_aic, calculate_bic

# All functions used for comparing a set of candidate models and
# ranking them take the same base arguments
# base - the base model
# candidates - a list of candidate models
# returns a list which is a subset of the candidates with the best ranked at the beginning
# models can be removed from candidates if they don't pass some criteria


def ofv(base, candidates, cutoff=3.84, rank_by_not_worse=False):
    return rank_models('ofv', base, candidates, cutoff, rank_by_not_worse=rank_by_not_worse)


def aic(base, candidates, cutoff=None):
    return rank_models('aic', base, candidates, cutoff)


def bic(base, candidates, cutoff=None, bic_type='mixed'):
    return rank_models('bic', base, candidates, cutoff, bic_type=bic_type)


def create_diff_dict(rank_type, base, candidates, bic_type='mixed'):
    delta_dict = dict()
    for model in candidates:
        if base.modelfit_results is not None and model.modelfit_results is not None:
            if rank_type == 'aic':
                delta = calculate_aic(base) - calculate_aic(model)
            elif rank_type == 'bic':
                if not bic_type:
                    bic_type = 'mixed'
                delta = calculate_bic(base, bic_type) - calculate_bic(model, bic_type)
            else:
                delta = base.modelfit_results.ofv - model.modelfit_results.ofv
        else:
            delta = np.nan
        delta_dict[model.name] = delta
    return delta_dict


def rank_models(
    rank_type, base, candidates, cutoff=None, rank_by_not_worse=False, bic_type='mixed'
):
    delta_dict = create_diff_dict(rank_type, base, candidates, bic_type)
    if cutoff is not None:
        if rank_type == 'ofv' and rank_by_not_worse:
            cutoff = -cutoff
        filtered = [model for model in candidates if delta_dict[model.name] >= cutoff]
    else:
        filtered = [model for model in candidates if model.modelfit_results is not None]

    def fn(model):
        if rank_type == 'aic':
            return calculate_aic(model)
        elif rank_type == 'bic':
            bic = calculate_bic(model, bic_type)
            return bic
        else:
            return getattr(model.modelfit_results, rank_type)

    srtd = sorted(filtered, key=fn)
    return srtd



from pharmpy.modeling.block_rvs import create_joint_distribution, split_joint_distribution
from pharmpy.modeling.common import (
    bump_model_number,
    convert_model,
    copy_model,
    fix_parameters,
    fix_parameters_to,
    generate_model_code,
    get_config_path,
    get_model_covariates,
    load_example_model,
    print_model_code,
    print_model_symbols,
    read_model,
    read_model_from_database,
    read_model_from_string,
    remove_unused_parameters_and_rvs,
    set_initial_estimates,
    set_name,
    unfix_parameters,
    unfix_parameters_to,
    write_model,
)
from pharmpy.modeling.covariate_effect import add_covariate_effect
from pharmpy.modeling.error import (
    has_additive_error_model,
    has_combined_error_model,
    has_proportional_error_model,
    remove_error_model,
    set_additive_error_model,
    set_combined_error_model,
    set_dtbs_error_model,
    set_proportional_error_model,
    set_time_varying_error_model,
    set_weighted_error_model,
    use_thetas_for_error_stdev,
)
from pharmpy.modeling.estimation_steps import (
    add_covariance_step,
    add_estimation_step,
    append_estimation_step_options,
    remove_covariance_step,
    remove_estimation_step,
    set_estimation_step,
    set_evaluation_step,
)
from pharmpy.modeling.eta_additions import add_iiv, add_iov, add_pk_iiv
from pharmpy.modeling.eta_transformations import (
    transform_etas_boxcox,
    transform_etas_john_draper,
    transform_etas_tdist,
)
from pharmpy.modeling.iiv_on_ruv import set_iiv_on_ruv
from pharmpy.modeling.math import (
    calculate_corr_from_cov,
    calculate_corr_from_inf,
    calculate_cov_from_corrse,
    calculate_cov_from_inf,
    calculate_inf_from_corrse,
    calculate_inf_from_cov,
    calculate_se_from_cov,
    calculate_se_from_inf,
)
from pharmpy.modeling.ml import (
    predict_influential_individuals,
    predict_influential_outliers,
    predict_outliers,
)
from pharmpy.modeling.odes import (
    add_individual_parameter,
    add_lag_time,
    add_peripheral_compartment,
    has_first_order_elimination,
    has_michaelis_menten_elimination,
    has_mixed_mm_fo_elimination,
    has_zero_order_absorption,
    has_zero_order_elimination,
    remove_lag_time,
    remove_peripheral_compartment,
    set_bolus_absorption,
    set_first_order_absorption,
    set_first_order_elimination,
    set_michaelis_menten_elimination,
    set_mixed_mm_fo_elimination,
    set_ode_solver,
    set_peripheral_compartments,
    set_seq_zo_fo_absorption,
    set_transit_compartments,
    set_zero_order_absorption,
    set_zero_order_elimination,
)
from pharmpy.modeling.parameter_sampling import (
    create_rng,
    sample_individual_estimates,
    sample_parameters_from_covariance_matrix,
    sample_parameters_uniformly,
)
from pharmpy.modeling.power_on_ruv import set_power_on_ruv
from pharmpy.modeling.remove_iiv import remove_iiv
from pharmpy.modeling.remove_iov import remove_iov
from pharmpy.modeling.run import create_results, fit, read_results, run_tool
from pharmpy.modeling.update_inits import update_inits
from pharmpy.modeling.write_csv import write_csv

from .allometry import add_allometry
from .amd import run_amd, run_iiv
from .data import (
    add_time_after_dose,
    check_dataset,
    drop_columns,
    drop_dropped_columns,
    expand_additional_doses,
    get_baselines,
    get_concentration_parameters_from_data,
    get_covariate_baselines,
    get_doseid,
    get_doses,
    get_ids,
    get_mdv,
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
    get_observations,
    list_time_varying_covariates,
    remove_loq_data,
    translate_nmtran_time,
    undrop_columns,
)
from .evaluation import (
    evaluate_epsilon_gradient,
    evaluate_eta_gradient,
    evaluate_expression,
    evaluate_individual_prediction,
    evaluate_population_prediction,
    evaluate_weighted_residuals,
)
from .expressions import (
    calculate_epsilon_gradient_expression,
    calculate_eta_gradient_expression,
    create_symbol,
    get_individual_prediction_expression,
    get_observation_expression,
    get_population_prediction_expression,
    mu_reference_model,
    simplify_expression,
)
from .iterators import omit_data, resample_data
from .plots import plot_individual_predictions, plot_iofv_vs_iofv
from .reporting import create_report
from .results import (
    calculate_aic,
    calculate_bic,
    calculate_eta_shrinkage,
    calculate_individual_parameter_statistics,
    calculate_individual_shrinkage,
    calculate_pk_parameters_statistics,
    check_high_correlations,
    check_parameters_near_bounds,
    print_fit_summary,
    summarize_modelfit_results,
    write_results,
)
from .units import get_unit_of

# Remember to sort __all__ alphabetically for order in documentation
__all__ = [
    'add_allometry',
    'add_covariance_step',
    'add_covariate_effect',
    'add_estimation_step',
    'add_iiv',
    'add_individual_parameter',
    'add_iov',
    'add_lag_time',
    'add_peripheral_compartment',
    'add_pk_iiv',
    'add_time_after_dose',
    'append_estimation_step_options',
    'bump_model_number',
    'calculate_aic',
    'calculate_bic',
    'calculate_corr_from_cov',
    'calculate_corr_from_inf',
    'calculate_cov_from_corrse',
    'calculate_cov_from_inf',
    'calculate_epsilon_gradient_expression',
    'calculate_eta_gradient_expression',
    'calculate_eta_shrinkage',
    'calculate_individual_parameter_statistics',
    'calculate_individual_shrinkage',
    'calculate_inf_from_corrse',
    'calculate_inf_from_cov',
    'calculate_pk_parameters_statistics',
    'calculate_se_from_cov',
    'calculate_se_from_inf',
    'check_dataset',
    'check_high_correlations',
    'check_parameters_near_bounds',
    'convert_model',
    'copy_model',
    'create_joint_distribution',
    'create_report',
    'create_results',
    'create_rng',
    'create_symbol',
    'drop_columns',
    'drop_dropped_columns',
    'evaluate_epsilon_gradient',
    'evaluate_eta_gradient',
    'evaluate_expression',
    'evaluate_individual_prediction',
    'evaluate_population_prediction',
    'evaluate_weighted_residuals',
    'expand_additional_doses',
    'fit',
    'fix_parameters',
    'fix_parameters_to',
    'get_baselines',
    'get_concentration_parameters_from_data',
    'get_covariate_baselines',
    'get_doses',
    'get_doseid',
    'get_ids',
    'get_individual_prediction_expression',
    'get_number_of_individuals',
    'get_number_of_observations',
    'get_number_of_observations_per_individual',
    'generate_model_code',
    'get_config_path',
    'get_mdv',
    'get_model_covariates',
    'get_observations',
    'get_observation_expression',
    'get_population_prediction_expression',
    'get_unit_of',
    'has_additive_error_model',
    'has_combined_error_model',
    'has_first_order_elimination',
    'has_michaelis_menten_elimination',
    'has_mixed_mm_fo_elimination',
    'has_proportional_error_model',
    'has_zero_order_absorption',
    'has_zero_order_elimination',
    'list_time_varying_covariates',
    'load_example_model',
    'mu_reference_model',
    'omit_data',
    'plot_individual_predictions',
    'plot_iofv_vs_iofv',
    'predict_influential_individuals',
    'predict_influential_outliers',
    'predict_outliers',
    'print_fit_summary',
    'print_model_code',
    'print_model_symbols',
    'read_model',
    'read_model_from_database',
    'read_model_from_string',
    'read_results',
    'remove_covariance_step',
    'remove_error_model',
    'remove_estimation_step',
    'remove_iiv',
    'remove_iov',
    'remove_lag_time',
    'remove_loq_data',
    'remove_peripheral_compartment',
    'remove_unused_parameters_and_rvs',
    'resample_data',
    'run_amd',
    'run_iiv',
    'run_tool',
    'sample_parameters_from_covariance_matrix',
    'sample_individual_estimates',
    'sample_parameters_uniformly',
    'set_additive_error_model',
    'set_bolus_absorption',
    'set_combined_error_model',
    'set_dtbs_error_model',
    'set_estimation_step',
    'set_evaluation_step',
    'set_first_order_absorption',
    'set_first_order_elimination',
    'set_iiv_on_ruv',
    'set_initial_estimates',
    'set_michaelis_menten_elimination',
    'set_mixed_mm_fo_elimination',
    'set_name',
    'set_ode_solver',
    'set_peripheral_compartments',
    'set_power_on_ruv',
    'set_proportional_error_model',
    'set_seq_zo_fo_absorption',
    'set_time_varying_error_model',
    'set_transit_compartments',
    'set_weighted_error_model',
    'set_zero_order_absorption',
    'set_zero_order_elimination',
    'simplify_expression',
    'split_joint_distribution',
    'summarize_modelfit_results',
    'transform_etas_boxcox',
    'transform_etas_john_draper',
    'transform_etas_tdist',
    'translate_nmtran_time',
    'unfix_parameters',
    'unfix_parameters_to',
    'update_inits',
    'use_thetas_for_error_stdev',
    'write_csv',
    'write_model',
    'write_results',
    'undrop_columns',
]



from pathlib import Path

from pharmpy.utils import TemporaryDirectory


def execute_workflow(workflow, dispatcher=None, database=None, path=None):
    """Execute workflow

    Parameters
    ----------
    dispatcher : ExecutionDispatcher
        Dispatcher to use. None for the default dispatcher
    database : ToolDatabase
        Tool database to use. None for the default Tool database.
    path : Path
        Path to use for database if applicable.

    Returns
    -------
    Results
        Results object created by workflow
    """
    if dispatcher is None:
        from pharmpy.workflows import default_dispatcher

        dispatcher = default_dispatcher
    if database is None:
        from pharmpy.workflows import default_tool_database

        database = default_tool_database(
            toolname=workflow.name, path=path
        )  # TODO: database -> tool_database

    # For all input models set new database and read in results
    original_input_models = []
    input_models = []
    for task in workflow.tasks:
        if task.has_input():
            new_inp = []
            from pharmpy.model import Model

            for inp in task.task_input:
                if isinstance(inp, Model):
                    original_input_models.append(inp)
                    inp.modelfit_results  # To read in the results
                    new_model = inp.copy()
                    new_model.dataset
                    new_model.database = database.model_database
                    new_inp.append(new_model)
                    input_models.append(new_model)
                else:
                    new_inp.append(inp)
            task.task_input = new_inp

    res = dispatcher.run(workflow)

    # Transfer files from tool model database to default model database
    for model in original_input_models:
        with TemporaryDirectory() as temppath:
            database.model_database.retrieve_local_files(model.name, temppath)
            for f in Path(temppath).glob('*'):
                # Copies all result files, copy model file if model does not have a file
                model_file = model.name + model.filename_extension
                if f.name != model_file or not (model.database.path / model_file).exists():
                    model.database.store_local_file(model, f)
        if isinstance(res, Model):
            # Special case to handle modelfit for generic models
            model.modelfit_results = res.modelfit_results
        else:
            # Set modelfit_results for local model objects
            try:
                model.read_modelfit_results()
            except NotImplementedError:
                pass

    from pharmpy.results import Results

    if isinstance(res, Results):
        database.store_results(res)
        if hasattr(res, 'rst_path'):
            from pharmpy.modeling.reporting import create_report

            create_report(res, database.path)

    return res


def split_common_options(d):
    """Split the dict into common options and other options

    Parameters
    ----------
    d : dict
        Dictionary of all options

    Returns
    -------
    Tuple of common options and other option dictionaries
    """
    execute_options = ['path']
    common_options = dict()
    other_options = dict()
    for key, value in d.items():
        if key in execute_options:
            common_options[key] = value
        else:
            other_options[key] = value
    return common_options, other_options



import re

from pharmpy.modeling import add_iiv, copy_model, create_joint_distribution, update_inits
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow

from .mfl import ModelFeatures


def exhaustive(mfl, add_iivs, iiv_as_fullblock, add_mdt_iiv):
    # TODO: rewrite using _create_model_workflow
    features = ModelFeatures(mfl)
    wf_search = Workflow()

    model_tasks = []
    model_features = dict()

    combinations = list(features.all_combinations())
    funcs = features.all_funcs()

    for i, combo in enumerate(combinations, 1):
        model_name = f'modelsearch_candidate{i}'

        task_copy = Task('copy', _copy, model_name)
        wf_search.add_task(task_copy)

        task_previous = task_copy
        for feat in combo:
            func = funcs[feat]
            task_function = Task(feat, func)
            wf_search.add_task(task_function, predecessors=task_previous)
            if add_iivs:
                task_add_iiv = Task(
                    'add_iivs', _add_iiv_to_func, feat, iiv_as_fullblock, add_mdt_iiv
                )
                wf_search.add_task(task_add_iiv, predecessors=task_function)
                task_previous = task_add_iiv
            else:
                task_previous = task_function

        wf_fit = create_fit_workflow(n=1)
        wf_search.insert_workflow(wf_fit, predecessors=task_previous)

        model_features[model_name] = tuple(combo)

        model_tasks += wf_fit.output_tasks

    return wf_search, model_tasks, model_features


def exhaustive_stepwise(mfl, add_iivs, iiv_as_fullblock, add_mdt_iiv):
    mfl_features = ModelFeatures(mfl)
    mfl_funcs = mfl_features.all_funcs()

    wf_search = Workflow()
    model_tasks = []
    model_features = dict()

    while True:
        no_of_trans = 0
        actions = _get_possible_actions(wf_search, mfl_features)
        for task_parent, feat_new in actions.items():
            for feat in feat_new:
                model_no = len(model_tasks) + 1
                model_name = f'modelsearch_candidate{model_no}'

                wf_create_model, task_function = _create_model_workflow(
                    model_name, feat, mfl_funcs[feat], add_iivs, iiv_as_fullblock, add_mdt_iiv
                )

                if task_parent:
                    wf_search.insert_workflow(wf_create_model, predecessors=[task_parent])
                else:
                    wf_search += wf_create_model

                model_tasks += wf_create_model.output_tasks
                features_previous = _get_previous_features(wf_search, task_function, mfl_funcs)
                model_features[model_name] = tuple(list(features_previous) + [feat])

                no_of_trans += 1
        if no_of_trans == 0:
            break

    return wf_search, model_tasks, model_features


def reduced_stepwise(mfl, add_iivs, iiv_as_fullblock, add_mdt_iiv):
    mfl_features = ModelFeatures(mfl)
    mfl_funcs = mfl_features.all_funcs()

    wf_search = Workflow()
    model_tasks = []
    model_features = dict()

    while True:
        no_of_trans = 0
        actions = _get_possible_actions(wf_search, mfl_features)
        if all(len(feat_new) > 0 for feat_new in actions.values()):
            groups = _find_same_model_groups(wf_search, mfl_funcs)
            if len(groups) > 1:
                for group in groups:
                    task_best_model = Task('choose_best_model', _get_best_model)
                    wf_search.add_task(task_best_model, predecessors=group)
                # Overwrite actions with new collector nodes
                actions = _get_possible_actions(wf_search, mfl_features)

        for task_parent, feat_new in actions.items():
            for feat in feat_new:
                model_no = len(model_tasks) + 1
                model_name = f'modelsearch_candidate{model_no}'

                wf_create_model, task_transformed = _create_model_workflow(
                    model_name, feat, mfl_funcs[feat], add_iivs, iiv_as_fullblock, add_mdt_iiv
                )

                if task_parent:
                    wf_search.insert_workflow(wf_create_model, predecessors=[task_parent])
                else:
                    wf_search += wf_create_model

                model_tasks += wf_create_model.output_tasks

                # FIXME: should this have same format as exhaustive?g
                model_features[model_name] = feat
                no_of_trans += 1
        if no_of_trans == 0:
            break

    return wf_search, model_tasks, model_features


def _find_same_model_groups(wf, mfl_funcs):
    tasks = wf.output_tasks
    tasks_removed = []
    all_groups = []
    for task_start in tasks:
        if id(task_start) in tasks_removed:
            continue
        else:
            tasks_removed += [id(task_start)]
        group = [task_start]
        features_previous = set(_get_previous_features(wf, task_start, mfl_funcs))
        for task in tasks:
            if (
                set(_get_previous_features(wf, task, mfl_funcs)) == features_previous
                and id(task) not in tasks_removed
            ):
                tasks_removed += [id(task)]
                group += [task]
        if len(group) > 1:
            all_groups += [group]
    return all_groups


def _get_best_model(*models):
    models_with_res = [model for model in models if model.modelfit_results]
    if models_with_res:
        return min(models, key=lambda x: x.modelfit_results.ofv)
    # FIXME: should be None, maybe dynamic workflows are needed
    return models[0]


def _get_possible_actions(wf, mfl_features):
    actions = dict()
    if wf.output_tasks:
        tasks = wf.output_tasks
    else:
        tasks = ['']
    for task in tasks:
        mfl_funcs = mfl_features.all_funcs()
        if task:
            feat_previous = _get_previous_features(wf, task, mfl_funcs)
        else:
            feat_previous = dict()

        trans_possible = [
            feat
            for feat, func in mfl_funcs.items()
            if _is_allowed(feat, func, feat_previous, mfl_features)
        ]

        actions[task] = trans_possible
    return actions


def _get_previous_features(wf, task, mfl_funcs):
    tasks_upstream = wf.get_upstream_tasks(task)
    tasks_upstream.reverse()
    features_previous = [task.name for task in tasks_upstream if task.name in mfl_funcs.keys()]
    return features_previous


def _create_model_workflow(model_name, feat, func, add_iivs, iiv_as_fullblock, add_mdt_iiv):
    wf_stepwise_step = Workflow()

    task_copy = Task('copy', _copy, model_name)
    wf_stepwise_step.add_task(task_copy)

    task_update_inits = Task('update_inits', _update_initial_estimates)
    wf_stepwise_step.add_task(task_update_inits, predecessors=task_copy)

    task_function = Task(feat, func)
    wf_stepwise_step.add_task(task_function, predecessors=task_update_inits)

    if add_iivs or add_mdt_iiv:
        task_add_iiv = Task('add_iivs', _add_iiv_to_func, feat, iiv_as_fullblock, add_mdt_iiv)
        wf_stepwise_step.add_task(task_add_iiv, predecessors=task_function)
        task_to_fit = task_add_iiv
    else:
        task_to_fit = task_function

    wf_fit = create_fit_workflow(n=1)
    wf_stepwise_step.insert_workflow(wf_fit, predecessors=task_to_fit)

    return wf_stepwise_step, task_function


def _is_allowed(feat_current, func_current, feat_previous, mfl_features):
    mfl_funcs = mfl_features.all_funcs()
    func_type = mfl_features.get_funcs_same_type(feat_current)
    # Check if current function is in previous transformations
    if feat_current in feat_previous:
        return False
    # Check if peripheral transformation is allowed
    if feat_current.startswith('PERIPHERALS'):
        peripheral_previous = [
            mfl_funcs[feat] for feat in feat_previous if feat.startswith('PERIPHERALS')
        ]
        return _is_allowed_peripheral(func_current, peripheral_previous, mfl_features)
    # Check if any functions of the same type has been used
    if any(mfl_funcs[feat] in func_type for feat in feat_previous):
        return False
    # No transformations have been made
    if not feat_previous:
        return True
    # Combinations to skip
    not_supported_combo = [
        ('ABSORPTION(ZO)', 'TRANSITS'),
        ('ABSORPTION(SEQ-ZO-FO)', 'TRANSITS'),
        ('ABSORPTION(SEQ-ZO-FO)', 'LAGTIME'),
        ('LAGTIME', 'TRANSITS'),
    ]
    for feat_1, feat_2 in not_supported_combo:
        if any(
            (feat_current.startswith(feat_1) and feat.startswith(feat_2))
            or (feat_current.startswith(feat_2) and feat.startswith(feat_1))
            for feat in feat_previous
        ):
            return False
    return True


def _is_allowed_peripheral(func_current, peripheral_previous, mfl_features):
    n_all = list(mfl_features.peripherals.args)
    n = func_current.keywords['n']
    if peripheral_previous:
        n_prev = [func.keywords['n'] for func in peripheral_previous]
    else:
        n_prev = []
    if not n_prev:
        if n == min(n_all):
            return True
        else:
            return False
    n_index = n_all.index(n)
    if n_index > 0 and n_all[n_index - 1] < n:
        return True
    return False


def _copy(name, model):
    model_copy = copy_model(model, name)
    return model_copy


def _update_initial_estimates(model):
    # FIXME: this should use dynamic workflows and not dispatch the next task
    try:
        update_inits(model, move_est_close_to_bounds=True)
    except ValueError:
        pass
    return model


def _add_iiv_to_func(feat, iiv_as_fullblock, add_mdt_iiv, model):
    # FIXME: Change to add_pk_iiv
    eta_dict = {
        'ABSORPTION(ZO)': ['MAT'],
        'ABSORPTION(SEQ-ZO-FO)': ['MAT', 'MDT'],
        'ELIMINATION(ZO)': ['CLMM', 'KM'],
        'ELIMINATION(MM)': ['CLMM', 'KM'],
        'ELIMINATION(MIX-FO-MM)': ['CLMM', 'KM'],
        'LAGTIME()': ['MDT'],
    }
    parameters = []
    if add_mdt_iiv:
        if feat == 'LAGTIME()' or feat.startswith('TRANSITS'):
            parameters = ['MDT']
        else:
            parameters = []
    else:
        if feat in eta_dict.keys():
            parameters = eta_dict[feat]
        elif feat.startswith('TRANSITS'):
            parameters = ['MDT']
        elif feat.startswith('PERIPHERALS'):
            no_of_peripherals = re.search(r'PERIPHERALS\((\d+)\)', feat).group(1)
            parameters = [f'VP{i}' for i in range(1, int(no_of_peripherals) + 1)] + [
                f'QP{i}' for i in range(1, int(no_of_peripherals) + 1)
            ]

    for param in parameters:
        assignment = model.statements.find_assignment(param)
        etas = {eta.name for eta in assignment.rhs_symbols}
        if etas.intersection(model.random_variables.etas.names):
            continue
        try:
            add_iiv(model, param, 'exp')
        except ValueError as e:
            if not str(e).startswith('Cannot insert parameter with already existing name'):
                raise
    if iiv_as_fullblock:
        create_joint_distribution(model)

    return model



import warnings

import pharmpy.workflows.dispatchers
from pharmpy.utils import TemporaryDirectory, TemporaryDirectoryChanger


def run(workflow):

    with TemporaryDirectory() as tempdirname:
        with TemporaryDirectoryChanger(tempdirname):
            workflow = workflow.as_dask_dict()

            if pharmpy.workflows.dispatchers.conf.dask_dispatcher:
                dask_dispatcher = pharmpy.workflows.dispatchers.conf.dask_dispatcher
            else:
                dask_dispatcher = 'distributed'

            if dask_dispatcher == 'threaded':
                from dask.threaded import get

                res = get(workflow, 'results')
            else:
                import tempfile

                import dask
                from dask.distributed import Client, LocalCluster

                dask.config.set({'temporary_directory': tempfile.gettempdir()})

                with warnings.catch_warnings():
                    # Catch deprecation warning from python 3.10 via tornado.
                    # Should be fixed with tornado 6.2
                    warnings.filterwarnings("ignore", message="There is no current event loop")
                    with LocalCluster(processes=False) as cluster:
                        with Client(cluster) as client:
                            print(client)
                            res = client.get(workflow, 'results')
    return res



import re
from pathlib import Path

from .option_record import OptionRecord


class TableRecord(OptionRecord):
    @property
    def path(self):
        file_option = self.option_pairs['FILE']
        return Path(file_option)

    @path.setter
    def path(self, value):
        self.set_option('FILE', value)

    @property
    def eta_derivatives(self):
        """List of numbers for etas whose derivative are requested"""
        return self._find_derivatives('G')

    @property
    def epsilon_derivatives(self):
        """List of numbers for epsilons whose derivateives are requested"""
        return self._find_derivatives('H')

    def _find_derivatives(self, ch):
        derivs = []
        regexp = ch + r'(\d+)1$'
        for key, value in self.all_options:
            m = re.match(regexp, key)
            if not m and value is not None:
                m = re.match(regexp, value)
            if m:
                n = m.group(1)
                derivs.append(int(n))
        return derivs



from pathlib import Path

import numpy as np
import pandas as pd

from .common import get_model_covariates
from .data import (
    get_ids,
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
)


def _create_dataset(model):
    idcol = model.datainfo.id_column.name
    nids = get_number_of_individuals(model)
    nobs = get_number_of_observations(model)
    nobsi = get_number_of_observations_per_individual(model)
    cwres = model.modelfit_results.residuals['CWRES']
    iofv = model.modelfit_results.individual_estimates
    npar = len(model.parameters)
    ofv = model.modelfit_results.ofv

    # Max ratio of abs(ETAi) and omegai
    variance_omegas = model.random_variables.etas.variance_parameters
    omega_estimates = np.sqrt(model.modelfit_results.parameter_estimates[variance_omegas])
    abs_ebes = model.modelfit_results.individual_estimates.abs()
    ebe_ratio = abs_ebes / list(omega_estimates)
    max_ebe_ratio = ebe_ratio.max(axis=1)

    # exp(OFVi / nobsi) / exp(OFV / nobs)
    iofv = model.modelfit_results.individual_ofv
    ofv_ratio = np.exp(iofv / nobsi) / np.exp(model.modelfit_results.ofv / nobs)

    # mean(ETA / OMEGA)
    cov = model.modelfit_results.individual_estimates_covariance
    etc_diag = np.sqrt(pd.DataFrame([np.diag(y) for y in cov], columns=cov.iloc[0].columns))
    etc_ratio = etc_diag / list(omega_estimates)
    mean_etc_ratio = etc_ratio.mean(axis=1)
    mean_etc_ratio.index = ofv_ratio.index

    # max((abs(indcov - mean(cov))) / sd(cov))
    cov_names = get_model_covariates(model, strings=True)
    if len(cov_names) > 0:
        covariates = model.dataset[cov_names + [idcol]].set_index(idcol)
        mean_covs = covariates.groupby(idcol).mean()
        maxcov = (abs(mean_covs - mean_covs.mean()) / mean_covs.std()).max(axis=1)
    else:
        maxcov = 0.0

    df = pd.DataFrame(
        {
            'nids': nids,
            'nobs': nobs,
            'nobs_subj': nobsi / (nobs / nids),
            'nobsi': nobsi,
            'ncovs': len(cov_names),
            'maxcov': maxcov,
            'max_cwres': abs(cwres).groupby(idcol).max(),
            'median_cwres': abs(cwres).groupby(idcol).median(),
            'max_ebe_ratio': max_ebe_ratio,
            'ofv_ratio': ofv_ratio,
            'etc_ratio': mean_etc_ratio,
            'iofv': iofv,
            'npar': npar,
            'ofv': ofv,
        }
    )
    return df


def predict_outliers(model, cutoff=3.0):
    """Predict outliers for a model using a machine learning model.

    See the :ref:`simeval <Individual OFV summary>` documentation for a definition of the `residual`

    Parameters
    ----------
    model : Model
        Pharmpy model
    cutoff : float
        Cutoff threshold for a residual singalling an outlier

    Returns
    -------
    pd.Dataframe
        Dataframe over the individuals with a `residual` column containing the raw predicted
        residuals and a `outlier` column with a boolean to tell whether the individual is
        an outlier or not.

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> predict_outliers(model)
        residual  outlier
    ID
    1  -0.281443    False
    2   1.057392    False
    3  -0.119105    False
    4  -0.846849    False
    5   0.600540    False
    6   1.014008    False
    7  -0.750734    False
    8   0.247175    False
    9   0.117002    False
    10 -0.835389    False
    11  3.529582     True
    12 -0.035670    False
    13  0.292333    False
    14  0.303278    False
    15 -0.565949    False
    16 -0.078192    False
    17 -0.291295    False
    18  2.466421    False
    19 -0.402343    False
    20 -0.699996    False
    21 -0.567987    False
    22 -0.526776    False
    23  0.303918    False
    24 -0.177588    False
    25  1.272142    False
    26 -0.390000    False
    27  0.775876    False
    28 -0.501528    False
    29 -0.700951    False
    30 -0.352599    False
    31  0.294196    False
    32  0.744014    False
    33 -0.215364    False
    34  0.208691    False
    35  1.713130    False
    36  0.300293    False
    37 -0.810736    False
    38  0.459877    False
    39 -0.675125    False
    40 -0.563835    False
    41 -0.526945    False
    42  4.449199     True
    43  0.720714    False
    44 -0.792248    False
    45 -0.860923    False
    46 -0.731858    False
    47 -0.247131    False
    48  1.894190    False
    49 -0.282737    False
    50 -0.153398    False
    51  1.200546    False
    52  0.902774    False
    53  0.586427    False
    54  0.183329    False
    55  1.036930    False
    56 -0.639868    False
    57 -0.765279    False
    58 -0.209665    False
    59 -0.225693    False

    See also
    --------
    predict_influential_individuals
    predict_influential_outliers

    """
    model_path = Path(__file__).parent.resolve() / 'ml_models' / 'outliers.tflite'
    data = _create_dataset(model)
    output = _predict_with_tflite(model_path, data)
    df = pd.DataFrame({'residual': output, 'outlier': output > cutoff}, index=get_ids(model))
    df.index.name = model.datainfo.id_column.name
    return df


def predict_influential_individuals(model, cutoff=3.84):
    """Predict influential individuals for a model using a machine learning model.

    Parameters
    ----------
    model : Model
        Pharmpy model
    cutoff : float
        Cutoff threshold for a dofv signalling an influential individual

    Returns
    -------
    pd.Dataframe
        Dataframe over the individuals with a `dofv` column containing the raw predicted
        delta-OFV and an `influential` column with a boolean to tell whether the individual is
        influential or not.

    See also
    --------
    predict_influential_outliers
    predict_outliers

    """

    model_path = Path(__file__).parent.resolve() / 'ml_models' / 'infinds.tflite'
    data = _create_dataset(model)
    output = _predict_with_tflite(model_path, data)
    df = pd.DataFrame({'dofv': output, 'influential': output > cutoff}, index=get_ids(model))
    df.index.name = model.datainfo.id_column.name
    return df


def predict_influential_outliers(model, outlier_cutoff=3, influential_cutoff=3.84):
    """Predict influential outliers for a model using a machine learning model.

    Parameters
    ----------
    model : Model
        Pharmpy model
    outlier_cutoff : float
        Cutoff threshold for a residual singalling an outlier
    influential_cutoff : float
        Cutoff threshold for a dofv signalling an influential individual

    Returns
    -------
    pd.Dataframe
        Dataframe over the individuals with a `outliers` and `dofv` columns containing the raw
        predictions and `influential`, `outlier` and `influential_outlier` boolean columns.

    See also
    --------
    predict_influential_individuals
    predict_outliers

    """

    outliers = predict_outliers(model, cutoff=outlier_cutoff)
    infinds = predict_influential_individuals(model, cutoff=influential_cutoff)
    df = pd.concat([outliers, infinds], axis=1)
    df['influential_outlier'] = df['outlier'] & df['influential']
    return df


def _predict_with_tflite(model_path, data):
    import tflite_runtime.interpreter as tflite

    interpreter = tflite.Interpreter(str(model_path))
    interpreter.allocate_tensors()
    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    npdata = data.astype('float32').values
    nrows = len(data)
    output = np.empty(nrows)

    for i in range(0, nrows):
        interpreter.set_tensor(input_details[0]['index'], npdata[i : (i + 1), :])
        interpreter.invoke()
        output_data = interpreter.get_tensor(output_details[0]['index'])
        output[i] = output_data

    return output



import shutil

from pharmpy import Model
from pharmpy.modeling import run_tool
from pharmpy.utils import TemporaryDirectoryChanger


def test_resmod(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)

        model = Model.create_model('pheno_real.mod')
        model.datainfo.path = tmp_path / 'pheno.dta'
        res = run_tool('resmod', model, groups=4, p_value=0.05, skip=[])
        assert res



from pathlib import Path


def write_csv(model, path=None, force=False):
    """Write dataset to a csv file

    Parameters
    ----------
    model : Model
        Model whose dataset to write to file
    path : Path
        Destination path. Default is to use original path with .csv suffix.
    force : bool
        Overwrite file with same path. Default is False.

    Returns
    -------
    Path
       path to the written file.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, write_csv
    >>> model = load_example_model("pheno")
    >>> write_csv(model, path="newdataset.csv")    # doctest: +SKIP
    newdataset.csv

    """
    import pharmpy.data

    if path is None:
        path = Path("")
    else:
        path = Path(path)
    if not path or path.is_dir():
        try:
            filename = f"{model.datainfo.path.name.with_suffix('.csv')}"
        except AttributeError:
            filename = f"{model.name + '.csv'}"
        path /= filename
    if not force and path.exists():
        raise FileExistsError(f'File at {path} already exists.')

    model.dataset.to_csv(path, na_rep=pharmpy.data.conf.na_rep, index=False)
    model.datainfo.path = path
    return path
